# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of all the available trainers."""
import dataclasses

from ml_collections.config_dict import config_dict

from optimus.trainer_lib import base_trainer


@dataclasses.dataclass(frozen=True)
class _Trainer:
  """Instance with artifact name, the corresponding class and hyperparameters.

  Attributes:
    name: A trainer class name.
    instance: A trainer class instance.
    hyperparameters: A trainer class hyperparameters.
  """

  name: str
  instance: type[base_trainer.BaseTrainer]
  hyperparameters: config_dict.ConfigDict


_ALL_TRAINERS = config_dict.ConfigDict(
    dict(
        base_trainer=_Trainer(
            "base_trainer",
            base_trainer.BaseTrainer,
            base_trainer.DEFAULT_HYPERPARAMETERS,
        )
    )
)


def get_trainer(
    *, trainer_name: str = "base_trainer"
) -> type[base_trainer.BaseTrainer]:
  """Maps the trainer name with the corresponding trainer class.

  Args:
    trainer_name: Trainer name.

  Returns:
    The requested trainer class.

  Raises:
    LookupError: An error when trying to access an unavailable trainer class.
  """
  if trainer_name not in _ALL_TRAINERS:
    raise LookupError(f"Unrecognized trainer class name: {trainer_name}")
  return _ALL_TRAINERS[trainer_name].instance


def get_trainer_hyperparameters(
    *, trainer_name: str = "base_trainer"
) -> config_dict.ConfigDict:
  """Maps the trainer name with the corresponding trainer hyperparameters.

  Args:
    trainer_name: Trainer name.

  Returns:
    The requested trainer class hyperparameters.

  Raises:
    LookupError: An error when trying to access an unavailable trainer class.
  """
  if trainer_name not in _ALL_TRAINERS:
    raise LookupError(f"Unrecognized trainer class name: {trainer_name}")
  return _ALL_TRAINERS[trainer_name].hyperparameters



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of all the available agents."""
import dataclasses
from typing import TypeVar

from ml_collections.config_dict import config_dict

from optimus.agent_lib import base_agent
from optimus.agent_lib import tabnet

_BaseAgentType = TypeVar("_BaseAgentType", bound=base_agent.BaseAgent)


@dataclasses.dataclass(frozen=True)
class _Agent:
  """Instance with and an agent class name, its instance and hyperparameters.

  Attributes:
    name: A name of an agent class.
    instance: An instance of an agent class.
    hyperparameters: Hyperparameters of an agent class.
  """

  name: str
  instance: object
  hyperparameters: config_dict.ConfigDict


_ALL_AGENTS = config_dict.ConfigDict(
    dict(
        tabnet=_Agent(
            "tabnet", tabnet.TabNetAgent, tabnet.DEFAULT_HYPERPARAMETERS
        )
    )
)


def get_agent(*, agent_name: str = "tabnet") -> _BaseAgentType:
  """Maps the agent name with the corresponding agent class.

  Args:
    agent_name: Agent name.

  Returns:
    The requested agent class.

  Raises:
    LookupError: An error when trying to access an unavailable agent class.
  """
  if agent_name not in _ALL_AGENTS:
    raise LookupError(f"Unrecognized agent class name: {agent_name}")
  return _ALL_AGENTS[agent_name].instance


def get_agent_hyperparameters(
    *, agent_name: str = "tabnet"
) -> config_dict.ConfigDict:
  """Maps the agent name with the corresponding agent hyperparameters.

  Args:
    agent_name: Agent name.

  Returns:
    The requested agent class hyperparameters.

  Raises:
    LookupError: An error when trying to access an unavailable agent class.
  """
  if agent_name not in _ALL_AGENTS:
    raise LookupError(f"Unrecognized agent class name: {agent_name}")
  return _ALL_AGENTS[agent_name].hyperparameters



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all trainers."""

import abc
import dataclasses
import functools
import itertools
from typing import Any, Callable, Mapping

from absl import logging
import flax
from flax.core import frozen_dict
import flax.linen as nn
from flax.metrics import tensorboard
from flax.training import common_utils
import jax
import jax.numpy as jnp
from ml_collections.config_dict import config_dict
import numpy as np
import optax
import orbax.checkpoint
import tensorflow as tf
import tqdm

from optimus.agent_lib import base_agent
from optimus.data_pipeline_lib import data_pipelines
from optimus.reward_lib import base_reward
from optimus.trainer_lib import losses
from optimus.trainer_lib import optimizers
from optimus.trainer_lib import schedules

DEFAULT_HYPERPARAMETERS = config_dict.ConfigDict(
    dict(
        train_steps=config_dict.placeholder(int),
        number_of_epochs=1,
        evaluation_frequency=1,
        evaluation_steps=1,
        gae_gamma=0.99,
        gae_lambda=0.95,
        clip_parameters=True,
        clip_parameters_coefficient=0.1,
        value_function_coefficient=0.5,
        entropy_coefficient=0.02,
        lambda_sparse=1e-3,
        training_rng_seed=-1,
        loss="tabnet_proximal_policy_optimization_loss",
        learning_rate_hyperparameters={
            "initial_value": 0.01,
            "schedule": "linear_schedule",
            "end_value": 0.0,
            "transition_steps": config_dict.placeholder(
                int
            ),  # Advised to be equal to train_steps
        },
        optimizer="adam",
        optimizer_hyperparameters={
            "b1": 0.9,
            "b2": 0.999,
            "eps": 1e-8,
            "eps_root": 0.0,
        },
        exploration_exploitation_hyperparameters={
            "schedule": "constant_schedule",
            "initial_value": 0.0,
        },
        checkpoint_frequency=1,
        checkpoints_kept=100,
        trainer_name="base_trainer",
    )
)


def calculate_advantages(
    previous_loop_result: float,
    current_array_element: jnp.ndarray,
    gae_gamma: float,
    gae_lambda: float,
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Calculates Generalized Advantage Estimation (GAE) advantages.

  See: https://arxiv.org/abs/1707.06347

  Args:
    previous_loop_result: The previous GAE value.
    current_array_element: The current element of stacked rewards, dones (masks)
      and values arrays.
    gae_gamma: A GAE gamma discounting factor.
    gae_lambda: A GAE lambda regularizing factor.

  Returns:
    An array with GAE advantages.

  Raises:
    ValueError: An error when length of the value array is too short.
  """
  reward, mask, value, next_value = current_array_element
  value_difference = gae_gamma * next_value * mask - value
  delta = reward + value_difference
  previous_loop_result = (
      delta
      + gae_gamma * gae_lambda * current_array_element[1] * previous_loop_result
  )
  return previous_loop_result, previous_loop_result


def calculate_trajectories(
    *, batch: Any, hyperparameters: config_dict.ConfigDict
) -> jnp.ndarray:
  """Processes agent's experience to calculate trajectories.

  Args:
    batch: A batch of data to be processed.
    hyperparameters: Experiment hyperparameters.

  Returns:
    A processed agent's experience described as trajectories.
  """
  values = batch["values"]
  modified_values = jnp.vstack((values, values[-1])).flatten()
  terminal_masks = batch["dones"].flatten()
  processed_terminal_masks = jnp.where(
      terminal_masks == 1.0, terminal_masks == 0.0, terminal_masks
  )
  stacked_arrays = jnp.stack(
      (
          batch["rewards"].flatten(),
          processed_terminal_masks,
          modified_values[:-1],
          modified_values[1:],
      ),
      axis=1,
  )
  partial_calculate_advantages = functools.partial(
      calculate_advantages,
      gae_gamma=hyperparameters.gae_gamma,
      gae_lambda=hyperparameters.gae_lambda,
  )
  generalized_advantage_estimation = 0.0
  _, unprocessed_advantages = jax.lax.scan(
      partial_calculate_advantages,
      generalized_advantage_estimation,
      stacked_arrays,
  )
  advantages = unprocessed_advantages.reshape(batch["states"].shape[0], 1)
  return jnp.concatenate(
      [
          batch["states"],
          batch["actions"],
          batch["log_probabilities"],
          advantages + values,
          advantages,
          batch["attentive_transformer_losses"],
      ],
      axis=1,
  )


def calculate_train_step_metrics(
    *,
    step: int,
    learning_rate: Callable[[int], float],
    exploration_exploitation_rate: Callable[[int], float],
    hyperparameters: config_dict.ConfigDict,
) -> dict[str, float]:
  """Returns a mapping between train step metrics and their values.

  Args:
    step: The current train step.
    learning_rate: A learning rate schedule.
    exploration_exploitation_rate: A exploration_exploitation_rate schedule.
    hyperparameters: Experiment hyperparameters.
  """
  alpha = (
      1.0 - step / hyperparameters.train_steps
      if hyperparameters.clip_parameters
      else 1.0
  )
  clip_parameters_coefficient = (
      hyperparameters.clip_parameters_coefficient * alpha
  )
  return {
      "step_learning_rate": learning_rate(step),
      "step_clip_parameters_coefficient": clip_parameters_coefficient,
      "step_exploration_exploitation_rate": exploration_exploitation_rate(step),
  }


def train_step(
    model_state: base_agent.BaseAgentState,
    *,
    exploration_exploitation_rate: float,
    clip_parameters_coefficient: float,
    batch: jnp.ndarray,
    loss_function: float | Callable[[float], float],
    hyperparameters: config_dict.ConfigDict,
) -> tuple[base_agent.BaseAgentState, jnp.ndarray]:
  """Returns an updated model state after a single training step.

  Args:
    model_state: The current model state.
    exploration_exploitation_rate: A value that controls how often an agent will
      choose an unoptimal action to explore.
    clip_parameters_coefficient: The clip parameters coefficient for a single
      step.
    batch: An input train batch.
    loss_function: The model's loss function, i.e.train_cost.
    hyperparameters: Experiment hyperparameters.
  """
  grad_function = jax.value_and_grad(loss_function, has_aux=True)
  (loss, updates), gradients = grad_function(
      model_state.params,
      model_state.batch_stats,
      model_state.apply_fn,
      batch=batch,
      clip_parameters_coefficient=clip_parameters_coefficient,
      hyperparameters=hyperparameters,
  )
  model_state = model_state.apply_gradients(
      grads=jax.lax.pmean(gradients, "batch")
  )
  model_state = model_state.replace(
      batch_stats=updates["batch_stats"],
      exploration_exploitation_epsilon=exploration_exploitation_rate,
  )
  loss = jax.lax.pmean(loss, "batch")
  return model_state, loss


def _sync_batch_statistics(
    *, model_state: base_agent.BaseAgentState
) -> base_agent.BaseAgentState:
  """Returns a model state with synced batch statistics across replicas.

  Args:
    model_state: The current model state.
  """
  cross_replica_mean = jax.pmap(lambda x: jax.lax.pmean(x, "x"), "x")
  return model_state.replace(
      batch_stats=cross_replica_mean(model_state.batch_stats)
  )


def evaluate(
    *,
    model_state: base_agent.BaseAgentState,
    evaluation_data: tf.data.Dataset,
    pmapped_evaluation_step: Any,
    evaluation_steps: int,
) -> dict[str, float]:
  """Calculates an evaluation reward.

  Args:
    model_state: The current model state.
    evaluation_data: A TensorFlow evaluation data pipeline.
    pmapped_evaluation_step: A pmapped evaluation step.
    evaluation_steps: A number of evaluation steps.

  Returns:
    A total evaluation reward.
  """
  evaluation_metrics = []
  for evaluation_batch in itertools.islice(
      evaluation_data, 0, evaluation_steps
  ):
    converted_evaluation_batch = jax.tree_util.tree_map(
        jnp.asarray, evaluation_batch
    )
    concatenated_evaluation_batch = jnp.concatenate(
        [
            converted_evaluation_batch["states"],
            converted_evaluation_batch["target_actions"],
        ],
        axis=1,
    )
    sharded_evaluation_batch = common_utils.shard(concatenated_evaluation_batch)
    evaluation_reward = pmapped_evaluation_step(
        agent_state=model_state,
        batch=sharded_evaluation_batch,
    )
    evaluation_metrics.append(
        dict(evaluation_reward=evaluation_reward.sum(keepdims=True))
    )
  return jax.tree_util.tree_map(
      jnp.sum, common_utils.get_metrics(evaluation_metrics)
  )


def evaluation_step(
    *,
    agent_state: base_agent.BaseAgentState,
    batch: jnp.ndarray,
    input_dimensions: int,
    agent: base_agent.BaseAgent,
    reward: base_reward.BaseReward,
) -> jnp.ndarray:
  """Calculates evaluation rewards given the selected and target actions.

  Args:
    agent_state: Current agent state.
    batch: An input batch with context for the agent to make predictions and
      target_actions to calculate evaluation rewards.
    input_dimensions: The number of dimesions in the context state.
    agent: A training agent.
    reward: A training reward.

  Returns:
    Evaluation rewards and booleans indicating if a sequence is complete.
  """
  states, target_actions = jnp.array_split(batch, [input_dimensions], axis=1)
  predicted_actions = agent.evaluate_predict(
      agent_state=agent_state, batch=states
  )
  return reward.calculate_evaluation_reward(
      predicted_actions=predicted_actions,
      target_actions=target_actions,
  )


def initialize_evaluation_data(
    *,
    hyperparameters: config_dict.ConfigDict,
    evaluation_dataset: tf.data.Dataset | None = None,
) -> tf.data.Dataset | None:
  """Returns an evaluation data pipeline.

  If BaseTrainer has evaluation_dataset loaded then evaluation_frequency
  and evaluation_steps hyperparameters are required for this method to work.

  Args:
    hyperparameters: Training hyperparameters.
    evaluation_dataset: A dataset with evaluation data with tensors of "states"
      and "dones".

  Raises:
    ValueError: An error if the constraints described above are violated.
  """
  if not evaluation_dataset:
    return None
  requirement_one = (
      "evaluation_frequency" in hyperparameters
      and hyperparameters.evaluation_frequency
  )
  requirement_two = (
      "evaluation_steps" in hyperparameters and hyperparameters.evaluation_steps
  )
  if not (requirement_one and requirement_two):
    raise ValueError(
        "The evaluation_frequency and evaluation_steps hyperparameters are "
        "required when evaluation_dataset is provided."
    )
  return evaluation_dataset


def initialize_model(
    *,
    initialization_rng_key: jax.Array,
    agent: base_agent.BaseAgent,
) -> tuple[nn.Module, frozen_dict.FrozenDict | Mapping[str, Any]]:
  """Initializes a Flax model.

  Args:
    initialization_rng_key: An RNG key for model initialization.
    agent: A training agent.

  Returns:
    Initialized FLAX model and initial model parameters.
  """
  dummy_initialization_batch = agent.get_dummy_inputs()
  flax_model = agent.flax_module
  initial_parameters = jax.jit(flax_model.init, static_argnames="train")(
      rngs={"params": initialization_rng_key},
      input_x=dummy_initialization_batch,
      train=False,
  )
  logging.info("Finished initializing the model.")
  logging.info("Model parameters: %s", initial_parameters)
  parameter_count = sum(
      x.size for x in jax.tree_util.tree_leaves(initial_parameters["params"])
  )
  logging.info("Number of model parameters: %s", parameter_count)
  return flax_model, initial_parameters


def initialize_schedule(
    *, hyperparameters: config_dict.ConfigDict
) -> Callable[[int], float] | float:
  """Returns a configured schedule.

  Args:
    hyperparameters: The schedule hyperparameters.
  """
  return schedules.get_schedule(schedule_name=hyperparameters.schedule)(
      hyperparameters
  )


def initialize_tensorboard_writer(
    *,
    hyperparameters: config_dict.ConfigDict,
) -> tensorboard.SummaryWriter:
  """Returns a TensorBoard SummaryWriter.

  Args:
    hyperparameters: Training hyperparameters.
  """
  tensorboard_writer = tensorboard.SummaryWriter(
      hyperparameters.artifact_directory
  )
  tensorboard_writer.hparams(hyperparameters.to_dict())
  return tensorboard_writer


def initialize_checkpointing(
    *,
    hyperparameters: config_dict.ConfigDict,
) -> tuple[orbax.checkpoint.CheckpointManager, int]:
  """Returns an Orbax CheckpointManager and the latest training step.

  Args:
    hyperparameters: Training hyperparameters.
  """
  checkpoint_manager_options = orbax.checkpoint.CheckpointManagerOptions(
      save_interval_steps=hyperparameters.checkpoint_frequency,
      max_to_keep=hyperparameters.checkpoints_kept,
      step_prefix="checkpoint",
      save_on_steps=(hyperparameters.train_steps,),
  )
  async_checkpointer = orbax.checkpoint.AsyncCheckpointer(
      orbax.checkpoint.PyTreeCheckpointHandler()
  )
  checkpoint_manager = orbax.checkpoint.CheckpointManager(
      directory=hyperparameters.checkpoint_directory,
      checkpointers=async_checkpointer,
      options=checkpoint_manager_options,
  )
  latest_training_step = checkpoint_manager.latest_step()
  if not latest_training_step:
    logging.info(
        "No checkpoint found under %s.",
        hyperparameters.checkpoint_directory,
    )
    latest_training_step = 0
  else:
    logging.info(
        "The latest saved checkpoint is: %s. The training loop will start"
        " from it.",
        latest_training_step,
    )
  logging.info("Training checkpointing is set up.")
  return checkpoint_manager, latest_training_step


def restore_model_state(
    *,
    checkpoint_manager: orbax.checkpoint.CheckpointManager,
    model_state: base_agent.BaseAgentState,
    hyperparameters: config_dict.ConfigDict,
    transforms: Any | None = None,
    mesh: jax.sharding.Mesh | None = None,
) -> base_agent.BaseAgentState:
  """Returns a restored TrainState.

  Args:
    checkpoint_manager: An Orbax CheckpointManager for the training procedure.
    model_state: A custom model state from the already initialized model.
    hyperparameters: Training hyperparameters.
    transforms: Transformations to apply to checkpoint found in `directory` (see
      orbax.checkpoint.transform_utils).
    mesh: Device mesh.
  """

  def _restore_arguments():
    if mesh is not None:
      return orbax.checkpoint.ArrayRestoreArgs(
          restore_type=jax.Array,
          mesh=mesh,
          mesh_axes=jax.sharding.PartitionSpec(),
      )
    return orbax.checkpoint.RestoreArgs(restore_type=np.ndarray)

  restore_arguments = jax.tree_util.tree_map(
      lambda _: _restore_arguments(), model_state
  )
  if transforms is None:
    transforms = {}
  transforms["step"] = orbax.checkpoint.Transform()
  restore_keyword_arguments = dict(
      restore_args=restore_arguments, transforms=transforms
  )
  return checkpoint_manager.restore(
      step=checkpoint_manager.latest_step(),
      restore_kwargs=restore_keyword_arguments,
      items=model_state,
      directory=hyperparameters.checkpoint_directory,
  )


def pmap_train_and_evaluation(
    *,
    agent: base_agent.BaseAgent,
    reward: base_reward.BaseReward,
    hyperparameters: config_dict.ConfigDict,
) -> tuple[Any, Any]:
  """Pmaps train and evaluation functions to allow for multi-device training.

  Args:
    agent: A training agent.
    reward: A training reward.
    hyperparameters: Training hyperparameters.

  Returns:
    Pmapped train and evaluation functions.
  """
  pmapped_train_step = jax.pmap(
      functools.partial(
          train_step,
          hyperparameters=hyperparameters,
          loss_function=losses.get_loss(loss_name=hyperparameters.loss),
      ),
      axis_name="batch",
  )
  pmapped_evaluation_step = jax.pmap(
      functools.partial(
          evaluation_step,
          input_dimensions=hyperparameters.input_dimensions,
          agent=agent,
          reward=reward,
      ),
      axis_name="batch",
  )
  logging.info("Training and evaluation functions are pmapped.")
  return pmapped_train_step, pmapped_evaluation_step


def is_evaluation_step(
    *, step: int, hyperparameters: config_dict.ConfigDict
) -> bool:
  """Returns an indicator whether to calculate evaluation metrics summary.

  Args:
    step: The latest training step from the Orbax Checkpoint Manager.
    hyperparameters: Training hyperparameters.
  """
  final_step = step == hyperparameters.train_steps - 1
  evaluation_metrics_summary_step = (
      step % hyperparameters.evaluation_frequency == 0
  )
  return final_step or evaluation_metrics_summary_step


def _process_replicated_model_state(
    *,
    model_state: base_agent.BaseAgentState,
) -> base_agent.BaseAgentState:
  """Returns an unreplicate model state with synced batch statistics.

  Args:
    model_state: The current model state.
  """
  model_state = _sync_batch_statistics(model_state=model_state)
  return jax.device_get(jax.tree_util.tree_map(lambda x: x[0], model_state))


def initialize_model_state(
    *,
    model: nn.Module,
    initial_parameters: frozen_dict.FrozenDict | Mapping[str, Any],
    optimizer: optax.GradientTransformation,
    exploration_exploitation_epsilon: float,
) -> base_agent.BaseAgentState:
  """Creates a custom model state.

  Args:
    model: An initialized Flax model.
    initial_parameters: Model parameters from an initialized model.
    optimizer: An initialized optimizer.
    exploration_exploitation_epsilon: A value that controls how often an agent
      will choose and an unoptimal action to explore.

  Returns:
    Custom model state.
  """
  model_state = base_agent.BaseAgentState.create(
      apply_fn=model.apply,
      params=initial_parameters["params"],
      tx=optimizer,
      batch_stats=initial_parameters["batch_stats"],
      exploration_exploitation_epsilon=exploration_exploitation_epsilon,
  )
  logging.info("Finished creating the model state.")
  return model_state


def initialize_model_state_for_prediction(
    *,
    agent: base_agent.BaseAgent,
    hyperparameters: config_dict.ConfigDict,
) -> base_agent.BaseAgentState:
  """Returns a model state initialized for prediction.

  Args:
    agent: A training agent.
    hyperparameters: Training hyperparameters.
  """
  rng_key = jax.random.PRNGKey(hyperparameters.training_rng_seed)
  initialization_rng_key, _ = jax.random.split(rng_key, 2)
  model, initial_parameters = initialize_model(
      initialization_rng_key=initialization_rng_key,
      agent=agent,
  )
  learning_rate = initialize_schedule(
      hyperparameters=hyperparameters.learning_rate_hyperparameters
  )
  exploration_exploitation_rate = initialize_schedule(
      hyperparameters=hyperparameters.exploration_exploitation_hyperparameters
  )
  optimizer = optimizers.get_optimizer(
      optimizer_name=hyperparameters.optimizer
  )(learning_rate, **hyperparameters.optimizer_hyperparameters)
  checkpoint_manager, train_loop_start_step = initialize_checkpointing(
      hyperparameters=hyperparameters
  )
  model_state = initialize_model_state(
      model=model,
      initial_parameters=initial_parameters,
      optimizer=optimizer,
      exploration_exploitation_epsilon=exploration_exploitation_rate(
          train_loop_start_step
      ),
  )
  if train_loop_start_step != 0:
    model_state = restore_model_state(
        checkpoint_manager=checkpoint_manager,
        model_state=model_state,
        hyperparameters=hyperparameters,
    )
  return model_state


@dataclasses.dataclass(frozen=True)
class TrainingBuildingBlocks:
  """Instance with and an agent class name, its instance and hyperparameters.

  Attributes:
    train_rng_key: A PRNG key for training repeatability.
    learning_rate: A learning rate schedule.
    exploration_exploitation_rate: A value that controls how often an agent will
      choose and an unoptimal action to explore.
    tensorboard_writer: The TensorBoard Summary Writer of the training
      procedure.
    train_loop_start_step: The latest training step from the Orbax Checkpoint
      Manager.
    checkpoint_manager: An Orbax CheckpointManager for the training procedure.
    pmapped_train_step: A pmapped train function.
    pmapped_evaluation_step: A pmapped evaluation function.
    model_state: An unreplicated model state.
    replicated_model_state: A replicated model state.
  """

  train_rng_key: jax.Array
  learning_rate: Callable[[int], float] | float
  exploration_exploitation_rate: Callable[[int], float] | float
  tensorboard_writer: tensorboard.SummaryWriter
  train_loop_start_step: int
  checkpoint_manager: orbax.checkpoint.CheckpointManager
  pmapped_train_step: Any
  pmapped_evaluation_step: Any
  model_state: base_agent.BaseAgentState
  replicated_model_state: base_agent.BaseAgentState

  @classmethod
  def prepare(
      cls,
      *,
      agent: base_agent.BaseAgent,
      reward: base_reward.BaseReward,
      hyperparameters: config_dict.ConfigDict,
  ) -> "TrainingBuildingBlocks":
    """Returns building blocks necessary for the training procedure.

    Args:
      agent: A training agent.
      reward: A training reward.
      hyperparameters: Training hyperparameters.
    """
    rng_key = jax.random.PRNGKey(hyperparameters.training_rng_seed)
    initialization_rng_key, train_rng_key = jax.random.split(rng_key, 2)
    model, initial_parameters = initialize_model(
        initialization_rng_key=initialization_rng_key,
        agent=agent,
    )
    learning_rate = initialize_schedule(
        hyperparameters=hyperparameters.learning_rate_hyperparameters
    )
    exploration_exploitation_rate = initialize_schedule(
        hyperparameters=hyperparameters.exploration_exploitation_hyperparameters
    )
    optimizer = optimizers.get_optimizer(
        optimizer_name=hyperparameters.optimizer
    )(learning_rate, **hyperparameters.optimizer_hyperparameters)
    tensorboard_writer = initialize_tensorboard_writer(
        hyperparameters=hyperparameters
    )
    checkpoint_manager, train_loop_start_step = initialize_checkpointing(
        hyperparameters=hyperparameters
    )
    model_state = initialize_model_state(
        model=model,
        initial_parameters=initial_parameters,
        optimizer=optimizer,
        exploration_exploitation_epsilon=exploration_exploitation_rate(
            train_loop_start_step
        ),
    )
    if train_loop_start_step != 0:
      model_state = restore_model_state(
          checkpoint_manager=checkpoint_manager,
          model_state=model_state,
          hyperparameters=hyperparameters,
      )
    pmapped_train_step, pmapped_evaluation_step = pmap_train_and_evaluation(
        hyperparameters=hyperparameters,
        agent=agent,
        reward=reward,
    )
    replicated_model_state = flax.jax_utils.replicate(model_state)
    return cls(
        train_rng_key=train_rng_key,
        learning_rate=learning_rate,
        exploration_exploitation_rate=exploration_exploitation_rate,
        tensorboard_writer=tensorboard_writer,
        train_loop_start_step=train_loop_start_step,
        checkpoint_manager=checkpoint_manager,
        pmapped_train_step=pmapped_train_step,
        pmapped_evaluation_step=pmapped_evaluation_step,
        model_state=model_state,
        replicated_model_state=replicated_model_state,
    )


def evaluate_on_step(
    *,
    hyperparameters: config_dict.ConfigDict,
    evaluation_data: tf.data.Dataset,
    training_blocks: TrainingBuildingBlocks,
    replicated_model_state: base_agent.BaseAgentState,
    step: int,
) -> None:
  """Evaluates an agent and saves the results to TensorBoard.

  Args:
    hyperparameters: Training hyperparameters.
    evaluation_data: An evaluation data pipeline.
    training_blocks: Building blocks of the training procedure.
    replicated_model_state: A replicated model state.
    step: The latest training step from the Orbax Checkpoint Manager.
  """
  evaluation_replicated_model_state = _sync_batch_statistics(
      model_state=replicated_model_state
  )
  evaluation_metrics_summary = evaluate(
      model_state=evaluation_replicated_model_state,
      evaluation_data=evaluation_data,
      pmapped_evaluation_step=training_blocks.pmapped_evaluation_step,
      evaluation_steps=hyperparameters.evaluation_steps,
  )
  for k, v in evaluation_metrics_summary.items():
    training_blocks.tensorboard_writer.scalar(k, v, step)


def update_model_state(
    *,
    batch: jnp.ndarray,
    training_blocks: TrainingBuildingBlocks,
    hyperparameters: config_dict.ConfigDict,
    replicated_model_state: base_agent.BaseAgentState,
    train_step_metrics: dict[str, float],
    step: int,
) -> tuple[base_agent.BaseAgentState, dict[str, float]]:
  """Returns an updated model state with train step metrics.

  Args:
    batch: An input train batch.
    training_blocks: Building blocks of the training procedure.
    hyperparameters: Training hyperparameters.
    replicated_model_state: A replicated model state.
    train_step_metrics: A mapping between train step metrics and their values.
    step: The latest training step from the Orbax Checkpoint Manager.
  """
  step_exploration_exploitation_rate = flax.jax_utils.replicate(
      train_step_metrics["step_exploration_exploitation_rate"]
  )
  step_clip_parameters_coefficient = flax.jax_utils.replicate(
      train_step_metrics["step_clip_parameters_coefficient"]
  )
  for epoch in range(hyperparameters.number_of_epochs + 1):
    permuted_trajectories = jax.random.permutation(
        jax.random.fold_in(training_blocks.train_rng_key, step + epoch),
        batch,
    )
    replicated_model_state, train_loss = training_blocks.pmapped_train_step(
        replicated_model_state,
        exploration_exploitation_rate=step_exploration_exploitation_rate,
        clip_parameters_coefficient=step_clip_parameters_coefficient,
        batch=common_utils.shard(permuted_trajectories),
    )
  train_step_metrics.update({"step_loss": train_loss.mean().item()})
  return replicated_model_state, train_step_metrics


def run_training_step(
    *,
    step: int,
    batch: jnp.ndarray,
    hyperparameters: config_dict.ConfigDict,
    replicated_model_state: base_agent.BaseAgentState,
    evaluation_data: tf.data.Dataset | None,
    training_blocks: TrainingBuildingBlocks,
) -> base_agent.BaseAgentState:
  """Returns an updated model state with train step metrics.

  Args:
    step: The latest training step from the Orbax Checkpoint Manager.
    batch: An input train batch.
    hyperparameters: Training hyperparameters.
    replicated_model_state: A replicated model state.
    evaluation_data: An evaluation data pipeline.
    training_blocks: Building blocks of the training procedure.
  """
  if evaluation_data and is_evaluation_step(
      step=step, hyperparameters=hyperparameters
  ):
    evaluate_on_step(
        hyperparameters=hyperparameters,
        evaluation_data=evaluation_data,
        training_blocks=training_blocks,
        replicated_model_state=replicated_model_state,
        step=step,
    )
  train_metrics_summary = calculate_train_step_metrics(
      step=step,
      learning_rate=training_blocks.learning_rate,
      exploration_exploitation_rate=training_blocks.exploration_exploitation_rate,
      hyperparameters=hyperparameters,
  )
  replicated_model_state, train_metrics_summary = update_model_state(
      batch=batch,
      training_blocks=training_blocks,
      hyperparameters=hyperparameters,
      replicated_model_state=replicated_model_state,
      train_step_metrics=train_metrics_summary,
      step=step,
  )
  for k, v in train_metrics_summary.items():
    training_blocks.tensorboard_writer.scalar(k, v, step)
  training_blocks.checkpoint_manager.save(
      step=step,
      items=_process_replicated_model_state(model_state=replicated_model_state),
  )
  training_blocks.checkpoint_manager.wait_until_finished()
  logging.info("Finished training step %d.", step)
  return replicated_model_state


def map_train_data(
    *,
    train_data: jnp.ndarray,
    hyperparameters: config_dict.ConfigDict,
) -> Mapping[str, jnp.ndarray]:
  """Returns mapped train arrays for the pretraining procedure.

  Args:
    train_data: An array with train data.
    hyperparameters: Training hyperparameters.
  """
  input_dimensions = hyperparameters.input_dimensions
  return dict(
      states=train_data[:, :input_dimensions],
      actions=train_data[:, input_dimensions : input_dimensions + 1],
      rewards=train_data[:, input_dimensions + 1 : input_dimensions + 2],
      values=train_data[:, input_dimensions + 2 : input_dimensions + 3],
      log_probabilities=train_data[
          :, input_dimensions + 3 : input_dimensions + 4
      ],
      dones=train_data[:, input_dimensions + 4 : input_dimensions + 5],
      attentive_transformer_losses=train_data[
          :, input_dimensions + 5 : input_dimensions + 6
      ],
  )


class BaseTrainer(metaclass=abc.ABCMeta):
  """Abstract parent class for all trainer classes.

  Attributes:
    agent: A training agent.
    reward: A training reward.
    hyperparameters: Training hyperparameters.
  """

  def __init__(
      self,
      *,
      agent: base_agent.BaseAgent,
      reward: base_reward.BaseReward,
      hyperparameters: config_dict.ConfigDict,
  ) -> None:
    """Initializes the Base Trainer class.

    Args:
      agent: A training agent.
      reward: A training reward.
      hyperparameters: Training hyperparameters.
    """
    self.agent = agent
    self.reward = reward
    self.hyperparameters = hyperparameters

  def train(
      self,
      *,
      train_data: Mapping[str, tf.Tensor | jnp.ndarray],
      evaluation_dataset: tf.data.Dataset | None = None,
      pretrain: bool = False,
      pretraining_blocks: TrainingBuildingBlocks | None = None,
      pretrain_model_state: base_agent.BaseAgentState | None = None,
      pretrain_step: int | None = None,
  ) -> base_agent.BaseAgentState:
    """Returns a model state after an end-to-end model training.

    Args:
      train_data: A batch with train data with collected experiences containing
        tensors / arrays of "states", "actions", "rewards", "values",
        "log_probabilities", "dones" and "attentive_transformer_losses". The
        dimensions of the arrays should be (batch_size, input_dimensions) for
        the "state" array and (batch_size, 1) for all the other arrays.
      evaluation_dataset: A dataset with tensors where the first N columns
        contain "state" and the last column contain the "target_action". The
        shape of the evaluation_dataset should be (evaluation_batch_size,
        input_dimensions + 1).
      pretrain: An indicator if the training occurs during pretraining session.
      pretraining_blocks: A dataclass with training blocks. It's only to be used
        during pretraining.
      pretrain_model_state: A model state to replicate. It's only to be used
        during pretraining.
      pretrain_step: A pretrain step. It's only to be used during pretraining.
    """
    if pretrain:
      if not pretraining_blocks:
        raise ValueError(
            "Pretraining requires pretraining blocks to be provided."
        )
      if not pretrain_model_state:
        raise ValueError(
            "Pretraining requires pretrain model state to be provided."
        )
      if not pretrain_step:
        raise ValueError("Pretraining requires pretrain step to be provided.")
      training_blocks = pretraining_blocks
      replicated_model_state = flax.jax_utils.replicate(pretrain_model_state)
      step = pretrain_step
    else:
      training_blocks = TrainingBuildingBlocks.prepare(
          hyperparameters=self.hyperparameters,
          agent=self.agent,
          reward=self.reward,
      )
      replicated_model_state = training_blocks.replicated_model_state
      step = training_blocks.train_loop_start_step + 1
    evaluation_data = initialize_evaluation_data(
        evaluation_dataset=evaluation_dataset,
        hyperparameters=self.hyperparameters,
    )
    trajectories = calculate_trajectories(
        batch=jax.tree_util.tree_map(jnp.asarray, train_data),
        hyperparameters=self.hyperparameters,
    )
    logging.info("Starting learning from the passed experience.")
    replicated_model_state = run_training_step(
        step=step,
        batch=trajectories,
        hyperparameters=self.hyperparameters,
        replicated_model_state=replicated_model_state,
        evaluation_data=evaluation_data,
        training_blocks=training_blocks,
    )
    if not pretrain:
      training_blocks.tensorboard_writer.close()
    logging.info("Finished learning.")
    return _process_replicated_model_state(model_state=replicated_model_state)

  def pretrain(
      self,
      *,
      pretrain_data: jnp.ndarray,
      evaluation_data: jnp.ndarray | None = None,
  ) -> base_agent.BaseAgentState:
    """Returns a model state after an end-to-end model training.

    Args:
      pretrain_data: A dataset with arrays where the first column is prediction
        seeds and the remaining columns of the array are train data. The shape
        of the pretrain_dataset should be (total_number_of_train_observations, 1
        + input_dimensions).
      evaluation_data: A dataset with arrays where the first N columns contain
        "state" and the last column contain the "target_action". The shape of
        the evaluation_data should be (total_number_of_evaluation_observations,
        input_dimensions + 1).
    """
    data_pipeline = data_pipelines.get_data_pipeline(
        data_pipeline_name=self.hyperparameters.data_pipeline_name
    )(hyperparameters=self.hyperparameters)
    train_pipeline_iterator = iter(
        data_pipeline.pretrain_data_pipeline(
            pretrain_data=pretrain_data,
        )
    )
    evaluation_pipeline = data_pipeline.evaluation_data_pipeline(
        evaluation_data=evaluation_data,
    )
    pretraining_blocks = TrainingBuildingBlocks.prepare(
        hyperparameters=self.hyperparameters,
        agent=self.agent,
        reward=self.reward,
    )
    model_state = pretraining_blocks.model_state
    train_loop_start_step = pretraining_blocks.train_loop_start_step + 1
    logging.info("Starting pretraining.")
    for step in tqdm.trange(
        train_loop_start_step, self.hyperparameters.train_steps + 1
    ):
      partial_pretrain_predict = functools.partial(
          self.agent.pretrain_predict,
          agent_state=model_state,
          calculate_pretrain_reward=self.reward.calculate_pretrain_reward,
      )
      batch = jnp.asarray(next(train_pipeline_iterator))
      train_data = jax.vmap(partial_pretrain_predict)(
          prediction_seed=batch[:, :1], batch=batch[:, 1:]
      )
      mapped_train_data = map_train_data(
          train_data=train_data, hyperparameters=self.hyperparameters
      )
      pretraining_blocks.tensorboard_writer.scalar(
          "training_reward",
          mapped_train_data["rewards"].sum().item(),
          step,
      )
      model_state = self.train(
          train_data=mapped_train_data,
          evaluation_dataset=evaluation_pipeline,
          pretrain=True,
          pretraining_blocks=pretraining_blocks,
          pretrain_model_state=model_state,
          pretrain_step=step,
      )
      logging.info("Finished pretraining step %d.", step)
    pretraining_blocks.tensorboard_writer.close()
    logging.info("Finished pretraining.")
    return model_state



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformer-based tabular data learning architecture.

Copyright (c) 2019 DreamQuark

TabNet:
https://arxiv.org/abs/1908.07442

The implementation is a refactored code from:
https://github.com/dreamquark-ai/tabnet
"""
import functools
import math
from typing import Any

from flax.core import frozen_dict
import flax.linen as nn
import jax
import jax.numpy as jnp
from ml_collections.config_dict import config_dict

from optimus.agent_lib import base_agent

FrozenDict = frozen_dict.FrozenDict

DEFAULT_HYPERPARAMETERS = config_dict.ConfigDict(
    dict(
        attention_gamma=1.3,
        prediction_layer_dimension=8,
        attention_layer_dimension=8,
        successive_network_steps=3,
        categorical_embedding_dimensions=[3],
        independent_glu_layers=2,
        shared_glu_layers=2,
        epsilon=1e-15,
        virtual_batch_size=128,
        momentum=0.98,
        mask_type="sparsemax",
        shared_decoder_layers=1,
        independent_decoder_layers=1,
        model_data_type="float32",
        agent_name="tabnet",
        replace_nans_in_prediction=False,
    )
)


class EmbeddingGenerator(nn.Module):
  """Generates embeddings in the model.

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    categorical_dimensions: The number of dimensions in each of the categorical
      columns of the input table.
    categorical_indexes: Indices of categorical columns in the input table.
    categorical_embedding_dimensions: The size of the embedding of categorical
      features if int, all categorical features will have same embedding size if
      list of int, every corresponding feature will have specific size.
    model_data_type: Data type to use within the model.
    embedding_initializer: Embeddings initializer.
    post_embedding_dimension: The post embedding dimension.
  """

  input_dimensions: int
  categorical_dimensions: tuple[int, ...] | None = None
  categorical_indexes: tuple[int, ...] | None = None
  categorical_embedding_dimensions: tuple[int, ...] | None = None
  model_data_type: jnp.dtype = jnp.float32
  embedding_initializer: Any = nn.initializers.normal(stddev=1**0.5)

  @functools.cached_property
  def _sorted_categorical_indexes(self) -> tuple[int, ...]:
    """Returns the sorted_indexes."""
    if self.categorical_indexes:
      return tuple([
          self.categorical_indexes.index(i)
          for i in sorted(self.categorical_indexes)
      ])
    else:
      raise ValueError("There is no categorical_indexes to sort.")

  @functools.cached_property
  def _verified_categorical_embedding_dimensions(
      self,
  ) -> tuple[int, ...] | None:
    """Returns the verified_categorical_embedding_dimensions."""
    if len(self.categorical_embedding_dimensions) == 1:
      categorical_embedding_dimensions = (
          self.categorical_embedding_dimensions * len(self.categorical_indexes)
      )
    else:
      categorical_embedding_dimensions = self.categorical_embedding_dimensions
    if len(categorical_embedding_dimensions) != len(
        self.categorical_dimensions
    ):
      raise ValueError(
          "categorical_embedding_dimensions and categorical_dimensions must"
          " be lists of same length, got"
          f" {len(self.categorical_embedding_dimensions)} and"
          f" {len(self.categorical_dimensions)}"
      )
    return categorical_embedding_dimensions

  @functools.cached_property
  def post_embedding_dimension(self) -> int:
    """Returns the post embedding dimension."""
    if not self.categorical_dimensions and not self.categorical_indexes:
      return self.input_dimensions
    return (
        self.input_dimensions
        + sum(self._verified_categorical_embedding_dimensions)
        - len(self._verified_categorical_embedding_dimensions)
    )

  def setup(self) -> None:
    """Sets up the module definition when it's called.

    Raises:
      ValueError: An error when length of categorical_dimensions is not equal to
      categorical_indexes. Or when length of categorical_embedding_dimensions
      is not equal to categorical_dimensions.
    """
    if not self.categorical_dimensions and not self.categorical_indexes:
      self.skip_embedding = True
      return
    if not self.categorical_dimensions or not self.categorical_indexes:
      raise ValueError(
          "Both categorical_dimensions and categorical_indexes "
          "are required if there are categorical columns."
      )
    if len(self.categorical_dimensions) != len(self.categorical_indexes):
      raise ValueError(
          "The lists categorical_dimensions and categorical_indexes must have"
          " the same length."
      )
    self.skip_embedding = False
    categorical_dimensions = [
        self.categorical_dimensions[i] for i in self._sorted_categorical_indexes
    ]
    embedding = functools.partial(
        nn.Embed,
        embedding_init=self.embedding_initializer,
        dtype=self.model_data_type,
    )
    processed_categorical_embedding_dimensions = [
        self._verified_categorical_embedding_dimensions[i]
        for i in self._sorted_categorical_indexes
    ]
    self.embeddings = [
        embedding(
            num_embeddings=categorical_dimensions + 1,
            features=embedding_dimensions,
        )
        for categorical_dimensions, embedding_dimensions in zip(
            categorical_dimensions,
            processed_categorical_embedding_dimensions,
        )
    ]
    self.continuous_indexes = [
        0
        if i in self.categorical_indexes
        else 1
        for i in range(self.input_dimensions)
    ]

  @nn.compact
  def __call__(self, *, input_x: jnp.ndarray) -> jnp.ndarray:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array used in the embedding process.

    Returns:
      The output of the forward pass.

    Raises:
      ValueError: An error when cat_feat_counter is higher than the nuber of
        embedding layers.
    """
    if self.skip_embedding:
      return input_x
    columns = []
    categorical_feature_counter = 0
    for feature_initial_index, is_continuous in enumerate(
        self.continuous_indexes
    ):
      if is_continuous:
        output = jnp.array(
            input_x[:, feature_initial_index], self.model_data_type
        ).reshape(-1, 1)
        columns.append(output)
      else:
        if categorical_feature_counter > len(self.embeddings):
          raise ValueError(
              "There're more categorical features than embedding layers."
          )
        embedding_layer = self.embeddings[categorical_feature_counter]
        embedding_input = jnp.array(
            input_x[:, feature_initial_index], dtype=jnp.int32
        )
        embedding_output = embedding_layer(inputs=embedding_input)
        columns.append(embedding_output)
        categorical_feature_counter += 1
    return jnp.concatenate(columns, axis=1)


class RandomObfuscator(nn.Module):
  """Used to randomly obfuscate column information during pre-training.

  Attributes:
    pretraining_ratio: The ratio of data to randomly obfuscate.
  """

  pretraining_ratio: float = 0.2

  @nn.compact
  def __call__(
      self, *, input_x: jnp.ndarray
  ) -> tuple[jnp.ndarray, jnp.ndarray]:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array where random information will be obfuscated.

    Returns:
      The output of the forward pass.
    """
    bernoulli_rng = self.make_rng("bernoulli_rng")
    obfuscated_vars = jax.random.bernoulli(
        bernoulli_rng, self.pretraining_ratio * jnp.ones(input_x.shape)
    )
    masked_input = jnp.multiply(1 - obfuscated_vars, input_x)
    return masked_input, obfuscated_vars


class GhostBatchNormalization(nn.Module):
  """Implementation of Ghost Batch Normalization.

  It enables significant decrease in the generalization gap without
  increasing the number of updates. For more details see:
  https://arxiv.org/abs/1705.08741

  Attributes:
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    model_data_type: Data type to use within the model.
  """

  virtual_batch_size: int = 128
  momentum: float = 0.99
  model_data_type: jnp.dtype = jnp.float32

  @nn.compact
  def __call__(
      self, *, input_x: jnp.ndarray, train: bool = False
  ) -> jnp.ndarray:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array to be processed by the Global Batch Normalization
        layer.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    batch_norm = nn.BatchNorm(
        momentum=self.momentum,
        use_running_average=not train,
        dtype=self.model_data_type,
    )
    indices_or_sections = math.ceil(input_x.shape[0] / self.virtual_batch_size)
    chunks = jnp.split(input_x, indices_or_sections, 0)
    result = [batch_norm(x=x) for x in chunks]
    return jnp.concatenate(result, axis=0)


class GatedLinearUnitLayer(nn.Module):
  """Implementation of a Gated Linear Unit Layer.

  See:
  https://arxiv.org/abs/2002.05202

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    output_dimensions: The number of possible actions.
    fully_connected_layer: A fully-connected Dense layer.
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  output_dimensions: int
  fully_connected_layer: nn.Module | None = None
  virtual_batch_size: int = 128
  momentum: float = 0.98
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called."""
    gain_value = jnp.sqrt(
        (self.input_dimensions + 2 * self.output_dimensions)
        / jnp.sqrt(self.input_dimensions)
    )
    kernel_init = nn.initializers.variance_scaling(
        scale=gain_value, mode="fan_avg", distribution="truncated_normal"
    )
    if self.fully_connected_layer:
      self.fc_layer = self.fully_connected_layer
    else:
      self.fc_layer = nn.Dense(
          2 * self.output_dimensions,
          use_bias=False,
          kernel_init=kernel_init,
          dtype=self.model_data_type,
      )
    self.ghost_batch_norm = GhostBatchNormalization(
        virtual_batch_size=self.virtual_batch_size,
        momentum=self.momentum,
        model_data_type=self.model_data_type,
    )

  @nn.compact
  def __call__(
      self, *, input_x: jnp.ndarray, train: bool = False
  ) -> jnp.ndarray:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array to be precessed by GLU layer.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    fully_connected_layer_output = self.fc_layer(inputs=input_x)
    ghost_batch_norm_output = self.ghost_batch_norm(
        input_x=fully_connected_layer_output, train=train
    )
    return jnp.multiply(
        ghost_batch_norm_output[:, : self.output_dimensions],
        nn.sigmoid(ghost_batch_norm_output[:, self.output_dimensions :]),
    )


class GatedLinearUnitBlock(nn.Module):
  """Implementation of a Gated Linear Unit Block.

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    output_dimensions: The number of possible actions.
    shared_layers: Fully-connected Dense layers.
    number_of_glu_layers: A number of GLU layers.
    first: Indicates if the first layer of the block has no scale
      multiplication.
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  output_dimensions: int
  shared_layers: tuple[nn.Module, ...] | None = None
  number_of_glu_layers: int = 2
  first: bool = False
  virtual_batch_size: int = 128
  momentum: float = 0.98
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called."""

    if self.shared_layers:
      self.glu_layers = [
          GatedLinearUnitLayer(
              self.input_dimensions,
              self.output_dimensions,
              fully_connected_layer=self.shared_layers[i],
              virtual_batch_size=self.virtual_batch_size,
              momentum=self.momentum,
              model_data_type=self.model_data_type,
          )
          for i in range(self.number_of_glu_layers)
      ]
    else:
      self.glu_layers = [
          GatedLinearUnitLayer(
              self.input_dimensions,
              self.output_dimensions,
              fully_connected_layer=None,
              virtual_batch_size=self.virtual_batch_size,
              momentum=self.momentum,
              model_data_type=self.model_data_type,
          )
          for _ in range(self.number_of_glu_layers)
      ]


  @nn.compact
  def __call__(
      self, *, input_x: jnp.ndarray, train: bool = False
  ) -> jnp.ndarray:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array to be processed by the GLU Block.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    scale = jnp.sqrt(jnp.array([0.5]))
    if self.first:
      input_x = self.glu_layers[0](input_x=input_x, train=train)
      layers_left = range(1, self.number_of_glu_layers)
    else:
      layers_left = range(self.number_of_glu_layers)

    for glu_id in layers_left:
      input_x = jnp.add(
          input_x, self.glu_layers[glu_id](input_x=input_x, train=train)
      )
      input_x *= scale
    return input_x


class Identity(nn.Module):
  """Implementation of the identity layer."""

  @nn.compact
  def __call__(
      self, *, input_x: jnp.ndarray, train: bool = False
  ) -> jnp.ndarray:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array to be passed through the Identiy layer.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    del train
    return input_x


class FeatureTransformer(nn.Module):
  """Feature transformer layer.

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    output_dimensions: The number of possible actions.
    number_of_independent_glu_layers: A number of independent GLU layers.
    shared_layers: Fully-connected Dense layers.
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  output_dimensions: int
  number_of_independent_glu_layers: int
  shared_layers: tuple[nn.Module, ...] | None = None
  virtual_batch_size: int = 128
  momentum: float = 0.98
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called."""
    if not self.shared_layers:
      self.shared = Identity()
      is_first = True
    else:
      self.shared = GatedLinearUnitBlock(
          input_dimensions=self.input_dimensions,
          output_dimensions=self.output_dimensions,
          first=True,
          shared_layers=self.shared_layers,
          number_of_glu_layers=len(self.shared_layers),
          virtual_batch_size=self.virtual_batch_size,
          momentum=self.momentum,
          model_data_type=self.model_data_type,
      )
      is_first = False
    if self.number_of_independent_glu_layers == 0:
      self.specifics = Identity()
    else:
      spec_input_dim = (
          self.input_dimensions if is_first else self.output_dimensions
      )
      self.specifics = GatedLinearUnitBlock(
          input_dimensions=spec_input_dim,
          output_dimensions=self.output_dimensions,
          first=is_first,
          number_of_glu_layers=self.number_of_independent_glu_layers,
          virtual_batch_size=self.virtual_batch_size,
          momentum=self.momentum,
          model_data_type=self.model_data_type,
      )

  @nn.compact
  def __call__(
      self, *, input_x: jnp.ndarray, train: bool = False
  ) -> jnp.ndarray:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array to be processed by the FeatureTransformer layer.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    shared_layers_output = self.shared(input_x=input_x, train=train)
    return self.specifics(input_x=shared_layers_output, train=train)


def _reshape_to_broadcast(
    *, indexes: jnp.ndarray, shape: tuple[int, ...], axis: int
) -> jnp.ndarray:
  """Utility function for the Sparsemax activation.

  Args:
    indexes: Indices from a specified axis.
    shape: Unmodified array shape.
    axis: An axis for the computation.

  Returns:
    The reshaped array for broadcasting.
  """
  new_shape = [1] * len(shape)
  new_shape[axis] = shape[axis]
  return jnp.reshape(indexes, new_shape)


@functools.partial(jax.custom_jvp, nondiff_argnums=(1,))
@functools.partial(jax.jit, static_argnums=(1,))
def _sparsemax(input_x: jnp.ndarray, axis: int) -> jnp.ndarray:
  """Utility function for the Sparsemax activation.

  The function defines the custom differentiation rule for the Sparsemax
  activation.

  Args:
    input_x: An input to use int the computation.
    axis: An axis for the computation.

  Returns:
    An output transformed with the Sparsemax activation function.
  """
  idxs = jnp.arange(input_x.shape[axis]) + 1
  idxs = _reshape_to_broadcast(indexes=idxs, shape=input_x.shape, axis=axis)
  sorted_x = jnp.flip(jax.lax.sort(input_x, dimension=axis), axis=axis)# jnp-type
  cum = jnp.cumsum(sorted_x, axis=axis)
  filtered_cum = jnp.sum(
      jnp.where(1 + sorted_x * idxs > cum, 1, 0), axis=axis, keepdims=True
  )
  threshold = (
      jnp.take_along_axis(cum, filtered_cum - 1, axis=axis) - 1
  ) / filtered_cum
  return jnp.maximum(input_x - threshold, 0)


@_sparsemax.defjvp
@functools.partial(jax.jit, static_argnums=(0,))
def _sparsemax_jvp(
    axis: int, primals: jnp.ndarray, tangents: jnp.ndarray
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Utility function for the Sparsemax activation.

  Args:
    axis: An axis for the computation.
    primals: Inputs to a primal function of type a.
    tangents: Corresponding tanget inputs of type T(a).

  Returns:
    Transformed input and gradients.
  """
  input_x = primals[0]
  tangent_input_x = tangents[0]
  entmax_projection = _sparsemax(input_x, axis)
  auxiliary_mask = jnp.where(entmax_projection > 0, 1, 0)
  tangent_y = tangent_input_x * auxiliary_mask
  gradients = jnp.sum(tangent_y, axis=axis) / jnp.sum(auxiliary_mask, axis=axis)
  tangent_y = tangent_y - jnp.expand_dims(gradients, axis) * auxiliary_mask
  return entmax_projection, tangent_y


def sparsemax(*, input_x: jnp.ndarray, axis: int) -> jnp.ndarray:
  """Sparsemax axtivation function.

  See: https://arxiv.org/abs/1602.02068
  See: https://github.com/deep-spin/entmax-jax

  Args:
    input_x: An input to use in the computation.
    axis: An axis for the computation.

  Returns:
    An output transformed with the Sparsemax activation function.
  """
  return _sparsemax(input_x, axis)


class AttentiveTransformer(nn.Module):
  """Attentive transformer layer.

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    output_dimensions: The number of possible actions.
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  output_dimensions: int
  virtual_batch_size: int = 128
  momentum: float = 0.98
  mask_type: str = "sparsemax"
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called."""
    gain_value = jnp.sqrt(
        (self.input_dimensions + self.output_dimensions)
        / jnp.sqrt(4 * self.input_dimensions)
    )
    kernel_init = nn.initializers.variance_scaling(
        scale=gain_value, mode="fan_avg", distribution="truncated_normal"
    )
    self.fully_connected_layer = nn.Dense(
        self.output_dimensions,
        use_bias=False,
        kernel_init=kernel_init,
        dtype=self.model_data_type,
    )
    self.ghost_batch_norm_output = GhostBatchNormalization(
        virtual_batch_size=self.virtual_batch_size,
        momentum=self.momentum,
        model_data_type=self.model_data_type,
    )

  @nn.compact
  def __call__(
      self, *, priors: jnp.ndarray, input_x: jnp.ndarray, train: bool = False
  ) -> jnp.ndarray:
    """Forward pass implementation for the layer.

    Args:
      priors: An array with priors scale information.
      input_x: An input array to be processed by the AttentiveTransformer layer.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.

    Raises:
      NotImplementedError: An error when trying to use an unavailble mask type.
    """
    if self.mask_type != "sparsemax":
      raise NotImplementedError("Only sparsemax is implemented.")
    fully_connected_layer_output = self.fully_connected_layer(inputs=input_x)
    ghost_batch_norm_output = self.ghost_batch_norm_output(
        input_x=fully_connected_layer_output, train=train
    )
    multiplification_output = jnp.multiply(ghost_batch_norm_output, priors)
    return sparsemax(input_x=multiplification_output, axis=-1)


class TabNetEncoder(nn.Module):
  """TabNet Encoder layer.

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    output_dimensions: The number of possible actions.
    prediction_layer_dimension: Dimension of the prediction layer (usually
      between 4 and 64).
    attention_layer_dimension: Dimension of the attention layer (usually between
      4 and 64).
    successive_network_steps: Number of successive steps in the network (usually
      between 3 and 10).
    attention_gamma: Float above 1, scaling factor for attention updates
      (usually between 1.0 to 2.0).
    independent_glu_layers: A number of independent GLU layers in each GLU
      block.
    shared_glu_layers: A number of shared GLU layer in each GLU block.
    epsilon: a factor to avoid jnp.log(0).
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    mask_type: The masking function to use.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  output_dimensions: int | tuple[int, ...]
  prediction_layer_dimension: int = 8
  attention_layer_dimension: int = 8
  successive_network_steps: int = 3
  attention_gamma: float = 1.3
  independent_glu_layers: int = 2
  shared_glu_layers: int = 2
  epsilon: float = 1e-15
  virtual_batch_size: int = 128
  momentum: float = 0.98
  mask_type: str = "sparsemax"
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called."""

    if self.shared_glu_layers == 0:
      shared_feat_transform = ()
    else:
      shared_feat_transform = [
          nn.Dense(
              2
              * (
                  self.prediction_layer_dimension
                  + self.attention_layer_dimension
              ),
              use_bias=False,
              dtype=self.model_data_type,
          )
          for _ in range(self.shared_glu_layers)
      ]
      shared_feat_transform = tuple(shared_feat_transform)

    self.initial_splitter = FeatureTransformer(
        input_dimensions=self.input_dimensions,
        output_dimensions=self.prediction_layer_dimension
        + self.attention_layer_dimension,
        shared_layers=shared_feat_transform,
        number_of_independent_glu_layers=self.independent_glu_layers,
        virtual_batch_size=self.virtual_batch_size,
        momentum=self.momentum,
        model_data_type=self.model_data_type,
    )
    self.feat_transformers = [
        FeatureTransformer(
            input_dimensions=self.input_dimensions,
            output_dimensions=self.prediction_layer_dimension
            + self.attention_layer_dimension,
            shared_layers=shared_feat_transform,
            number_of_independent_glu_layers=self.independent_glu_layers,
            virtual_batch_size=self.virtual_batch_size,
            momentum=self.momentum,
            model_data_type=self.model_data_type,
        )
    ] * self.successive_network_steps
    self.att_transformers = [
        AttentiveTransformer(
            input_dimensions=self.attention_layer_dimension,
            output_dimensions=self.input_dimensions,
            virtual_batch_size=self.virtual_batch_size,
            momentum=self.momentum,
            mask_type=self.mask_type,
            model_data_type=self.model_data_type,
        )
    ] * self.successive_network_steps

  @nn.compact
  def __call__(# jnp-array
      self,
      *,
      input_x: jnp.ndarray,
      prior: jnp.ndarray = None,
      forward_masks: bool = False,
      train: bool = False,
  ) -> (
      tuple[list[jnp.ndarray], list[jnp.ndarray]]
      | tuple[jnp.ndarray, dict[int, jnp.ndarray]]
  ):
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array to be processed by the TabNetEncoder layer.
      prior: An array with prior scale information.
      forward_masks: An indicator if the forward_masks process should be called.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    batch_norm = nn.BatchNorm(
        momentum=0.99, use_running_average=not train, dtype=self.model_data_type
    )
    batch_normalization_output = batch_norm(x=input_x)
    if not forward_masks:
      if prior is None:
        prior = jnp.ones(
            batch_normalization_output.shape, dtype=self.model_data_type
        )
      attentive_transformer_loss = []
      feature_transformer_output = self.initial_splitter(
          input_x=batch_normalization_output, train=train
      )[:, self.prediction_layer_dimension :]
      steps_output = []
      for step in range(self.successive_network_steps):
        attentive_transformer_layer_output = self.att_transformers[step](
            priors=prior, input_x=feature_transformer_output, train=train
        )
        attentive_transformer_loss.append(
            jnp.mean(
                jnp.sum(
                    jnp.multiply(
                        attentive_transformer_layer_output,
                        jnp.log(
                            attentive_transformer_layer_output + self.epsilon
                        ),
                    ),
                    axis=1,
                )
            )
        )
        prior = jnp.multiply(
            self.attention_gamma - attentive_transformer_layer_output, prior
        )
        masked_x = jnp.multiply(
            attentive_transformer_layer_output, batch_normalization_output
        )
        output = self.feat_transformers[step](input_x=masked_x, train=train)
        post_activation_out = nn.relu(
            output[:, : self.prediction_layer_dimension]
        )
        steps_output.append(post_activation_out)
        feature_transformer_output = output[
            :, self.prediction_layer_dimension :
        ]

      return steps_output, attentive_transformer_loss

    prior = jnp.ones(
        batch_normalization_output.shape, dtype=self.model_data_type
    )
    attentive_transformer_explain = jnp.zeros(
        batch_normalization_output.shape, dtype=self.model_data_type
    )
    feature_transformer_output = self.initial_splitter(
        input_x=batch_normalization_output, train=train
    )[:, self.prediction_layer_dimension :]
    masks = {}

    for step in range(self.successive_network_steps):
      attentive_transformer_layer_output = self.att_transformers[step](
          priors=prior, input_x=feature_transformer_output
      )
      masks[step] = attentive_transformer_layer_output
      prior = jnp.multiply(
          self.attention_gamma - attentive_transformer_layer_output, prior
      )
      masked_x = jnp.multiply(
          attentive_transformer_layer_output, batch_normalization_output
      )
      output = self.feat_transformers[step](input_x=masked_x, train=train)
      post_activation_out = nn.relu(
          output[:, : self.prediction_layer_dimension]
      )
      step_importance = jnp.sum(post_activation_out, axis=1)
      attentive_transformer_explain += jnp.multiply(
          attentive_transformer_layer_output,
          jnp.expand_dims(step_importance, axis=1),
      )
      feature_transformer_output = output[:, self.prediction_layer_dimension :]
    return attentive_transformer_explain, masks


class TabNetDecoder(nn.Module):
  """TabNet Decoder layer.

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    prediction_layer_dimension: Dimension of the prediction layer (usually
      between 4 and 64).
    successive_network_steps: Number of successive steps in the network (usually
      between 3 and 10).
    independent_glu_layers: A number of independent GLU layer in each GLU block.
    shared_glu_layers: A number of shared GLU layers in each GLU block.
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  prediction_layer_dimension: int = 8
  successive_network_steps: int = 3
  independent_glu_layers: int = 1
  shared_glu_layers: int = 1
  virtual_batch_size: int = 128
  momentum: float = 0.98
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called."""
    if self.shared_glu_layers == 0:
      shared_feat_transform = ()
    else:
      shared_feat_transform = [
          nn.Dense(2 * self.prediction_layer_dimension, use_bias=False)
      ] * self.shared_glu_layers
      shared_feat_transform = tuple(shared_feat_transform)
    self.feat_transformers = [
        FeatureTransformer(
            input_dimensions=self.prediction_layer_dimension,
            output_dimensions=self.prediction_layer_dimension,
            shared_layers=shared_feat_transform,
            number_of_independent_glu_layers=self.independent_glu_layers,
            virtual_batch_size=self.virtual_batch_size,
            momentum=self.momentum,
            model_data_type=self.model_data_type,
        )
    ] * self.successive_network_steps
    gain_value = jnp.sqrt(
        (self.input_dimensions + self.prediction_layer_dimension)
        / jnp.sqrt(4 * self.input_dimensions)
    )
    kernel_init = nn.initializers.variance_scaling(
        scale=gain_value, mode="fan_avg", distribution="truncated_normal"
    )
    self.reconstruction_layer = nn.Dense(
        self.input_dimensions,
        use_bias=False,
        kernel_init=kernel_init,
        dtype=self.model_data_type,
    )

  @nn.compact
  def __call__(
      self, *, input_x: list[float], train: bool = False
  ) -> jnp.ndarray:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    result = 0.0
    for step_nb, step_output in enumerate(input_x):
      feature_transformer_output = self.feat_transformers[step_nb](# jnp-array
          input_x=step_output, train=train
      )
      result = jnp.add(result, feature_transformer_output)
    return self.reconstruction_layer(inputs=result)


class TabNetPretraining(nn.Module):
  """TabNet Pretraining layer.

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    pretraining_ratio: The ratio of data to randomly obfuscate. between 4 and
      64).
    attention_layer_dimension: Dimension of the attention layer (usually between
      4 and 64).
    successive_network_steps: Number of successive steps in the network (usually
      between 3 and 10).
    attention_gamma: Float above 1, scaling factor for attention updates
      (usually between 1.0 to 2.0).
    categorical_indexes: Indices of categorical columns in the input table.
    categorical_indexes: Indices of categorical columns in the input table.
    categorical_dimensions: The number of dimensions in each of the categorical
      columns of the input table.
    categorical_embedding_dimensions: The size of the embedding of categorical
      features if int, all categorical features will have same embedding size if
      list of int, every corresponding feature will have specific size.
    independent_glu_layers: A number of independent GLU layers in each GLU
      block.
    shared_glu_layers: A number of shared GLU layers in each GLU block.
    epsilon: a constant to avoid jnp.log(0).
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    mask_type: The masking function to use.
    shared_decoder_layers: A number of shared decoder layers in each GLU block
      in the decoder.
    independent_decoder_layers: A number of independent decoder layers in each
      GLU block in the decoder.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  pretraining_ratio: float = 0.2
  prediction_layer_dimension: int = 8
  attention_layer_dimension: int = 8
  successive_network_steps: int = 3
  attention_gamma: float = 1.3
  categorical_dimensions: tuple[int, ...] | None = None
  categorical_indexes: tuple[int, ...] | None = None
  categorical_embedding_dimensions: tuple[int, ...] | None = None
  independent_glu_layers: int = 2
  shared_glu_layers: int = 2
  epsilon: float = 1e-15
  virtual_batch_size: int = 128
  momentum: float = 0.98
  mask_type: str = "sparsemax"
  shared_decoder_layers: int = 1
  independent_decoder_layers: int = 1
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called.

    Raises:
      ValueError: An error when successive_network_steps is less or equal to 0.
        Or when independent_glu_layers and shared_glu_layers are equal 0.
    """
    if self.successive_network_steps <= 0:
      raise ValueError("successive_network_steps should be a positive integer.")
    if self.independent_glu_layers == 0 and self.shared_glu_layers == 0:
      raise ValueError(
          "shared_glu_layers and independent_glu_layers can not be both zero."
      )
    categorical_dimensions = self.categorical_dimensions or ()
    categorical_indexes = self.categorical_indexes or ()
    self.embedder = EmbeddingGenerator(
        self.input_dimensions,
        categorical_dimensions,
        categorical_indexes,
        self.categorical_embedding_dimensions,
        self.model_data_type,
    )
    self.post_embed_dim = self.embedder.post_embedding_dimension
    self.masker = RandomObfuscator(self.pretraining_ratio)
    self.encoder = TabNetEncoder(
        input_dimensions=self.post_embed_dim,
        output_dimensions=self.post_embed_dim,
        prediction_layer_dimension=self.prediction_layer_dimension,
        attention_layer_dimension=self.attention_layer_dimension,
        successive_network_steps=self.successive_network_steps,
        attention_gamma=self.attention_gamma,
        independent_glu_layers=self.independent_glu_layers,
        shared_glu_layers=self.shared_glu_layers,
        epsilon=self.epsilon,
        virtual_batch_size=self.virtual_batch_size,
        momentum=self.momentum,
        mask_type=self.mask_type,
        model_data_type=self.model_data_type,
    )
    self.decoder = TabNetDecoder(
        input_dimensions=self.post_embed_dim,
        prediction_layer_dimension=self.prediction_layer_dimension,
        successive_network_steps=self.successive_network_steps,
        independent_glu_layers=self.independent_decoder_layers,
        shared_glu_layers=self.shared_decoder_layers,
        virtual_batch_size=self.virtual_batch_size,
        momentum=self.momentum,
        model_data_type=self.model_data_type,
    )

  @nn.compact
  def __call__(
      self, *, input_x: jnp.ndarray, train: bool = False
  ) -> tuple[jnp.ndarray, ...]:
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array to be processed by the TabNetPretraining layer.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    embedded_x = self.embedder(input_x=input_x)
    if train:
      masked_x, obf_vars = self.masker(input_x=embedded_x)
      prior = 1 - obf_vars
      steps_out, _ = self.encoder(input_x=masked_x, prior=prior, train=train)
      result = self.decoder(input_x=steps_out, train=train)# jnp-array
      return result, embedded_x, obf_vars
    steps_out, _ = self.encoder(input_x=embedded_x, train=train)
    result = self.decoder(input_x=steps_out, train=train)# jnp-array
    return result, embedded_x, jnp.ones(embedded_x.shape)


class TabNetCoreTraining(nn.Module):
  """TabNet Core Training layer.

  The TabNet architecture used for training and NOT pretraining.

  Attributes:
    input_dimensions: The number of feature columns in the input table.
    output_dimensions: The number of possible actions.
    prediction_layer_dimension: Dimension of the prediction layer (usually
      between 4 and 64).
    attention_layer_dimension: Dimension of the attention layer (usually between
      4 and 64).
    successive_network_steps: Number of successive steps in the network (usually
      between 3 and 10).
    attention_gamma: Float above 1, scaling factor for attention updates
      (usually between 1.0 to 2.0).
    independent_glu_layers: A number of independent GLU layer in each GLU block.
    shared_glu_layers: A number of shared GLU layer in each GLU block.
    epsilon: A factor to avoid jnp.log(0).
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    mask_type: The masking function to use.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  output_dimensions: int | tuple[int, ...]
  prediction_layer_dimension: int = 8
  attention_layer_dimension: int = 8
  successive_network_steps: int = 3
  attention_gamma: float = 1.3
  independent_glu_layers: int = 2
  shared_glu_layers: int = 2
  epsilon: float = 1e-15
  virtual_batch_size: int = 128
  momentum: float = 0.98
  mask_type: str = "sparsemax"
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called."""

    self.encoder = TabNetEncoder(
        input_dimensions=self.input_dimensions,
        output_dimensions=self.output_dimensions,
        prediction_layer_dimension=self.prediction_layer_dimension,
        attention_layer_dimension=self.attention_layer_dimension,
        successive_network_steps=self.successive_network_steps,
        attention_gamma=self.attention_gamma,
        independent_glu_layers=self.independent_glu_layers,
        shared_glu_layers=self.shared_glu_layers,
        epsilon=self.epsilon,
        virtual_batch_size=self.virtual_batch_size,
        momentum=self.momentum,
        mask_type=self.mask_type,
        model_data_type=self.model_data_type,
    )

    if isinstance(self.output_dimensions, tuple):
      n_preds = [self.prediction_layer_dimension] * len(self.output_dimensions)
      gain_values = [
          jnp.sqrt((i + o) / jnp.sqrt(4 * i))
          for i, o in zip(n_preds, self.output_dimensions)
      ]
      kernel_inits = [
          nn.initializers.variance_scaling(
              scale=g, mode="fan_avg", distribution="truncated_normal"
          )
          for g in gain_values
      ]
      self.multi_task_mappings = [
          nn.Dense(o, use_bias=False, kernel_init=k, dtype=self.model_data_type)
          for o, k in zip(self.output_dimensions, kernel_inits)
      ]
    else:
      gain_value = jnp.sqrt(
          (self.prediction_layer_dimension + self.output_dimensions)
          / jnp.sqrt(4 * self.prediction_layer_dimension)
      )
      kernel_init = nn.initializers.variance_scaling(
          scale=gain_value, mode="fan_avg", distribution="truncated_normal"
      )
      self.final_mapping = nn.Dense(
          self.output_dimensions,
          use_bias=False,
          kernel_init=kernel_init,
          dtype=self.model_data_type,
      )
    self.value_layer = nn.Dense(
        features=1, name="value", dtype=self.model_data_type
    )

  @nn.compact
  def __call__(
      self,
      *,
      input_x: jnp.ndarray,
      forward_masks: bool = False,
      train: bool = False,
  ) -> (
      tuple[list[jnp.ndarray], list[jnp.ndarray]]
      | tuple[jnp.ndarray, dict[int, jnp.ndarray]]
      | tuple[jnp.ndarray, jnp.ndarray, float]
  ):
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array.
      forward_masks: An indicator if the forward_masks process should be called.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    if not forward_masks:
      steps_output, attentive_transformer_loss = self.encoder(
          input_x=input_x, train=train
      )
      result = jnp.sum(jnp.stack(steps_output, axis=0), axis=0)
      attentive_transformer_loss = jnp.mean(
          jnp.stack(attentive_transformer_loss, axis=0), axis=0
      )
      if isinstance(self.output_dimensions, tuple):
        logits = []
        for task_mapping in self.multi_task_mappings:
          logits.append(task_mapping(inputs=result))
        policy_log_probabilities = [nn.log_softmax(i) for i in logits]
      else:
        logits = self.final_mapping(inputs=result)
        policy_log_probabilities = nn.log_softmax(logits)
      value = self.value_layer(inputs=result)
      return (policy_log_probabilities, value, attentive_transformer_loss)# jnp-array
    return self.encoder(input_x=input_x, train=train, forward_masks=True)


class TabNet(nn.Module):
  """TabNet model.

  Attributes:
    output_dimensions: The number of possible actions.
    prediction_layer_dimension: Dimension of the prediction layer (usually
      between 4 and 64).
    attention_layer_dimension: Dimension of the attention layer (usually between
      4 and 64).
    successive_network_steps: Number of successive steps in the network (usually
      between 3 and
    categorical_indexes: Indices of categorical columns in the input table.
    categorical_dimensions: The number of dimensions in each of the categorical
    categorical_embedding_dimensions: The size of the embedding of categorical
      features if int, all categorical features will have same embedding size if
      list of int, every corresponding feature will have specific size.
    independent_glu_layers: A number of independent GLU layers in each GLU
      block.
    shared_glu_layers: A number of shared GLU layer in each GLU block.
    epsilon: A factor to avoid jnp.log(0).
    virtual_batch_size: Batch size for Ghost Batch Normalization.
    momentum: Momentum value for the BatchNorm layer.
    mask_type: The masking function to use.
    model_data_type: Data type to use within the model.
  """

  input_dimensions: int
  output_dimensions: int | tuple[int, ...]
  prediction_layer_dimension: int = 8
  attention_layer_dimension: int = 8
  successive_network_steps: int = 3
  attention_gamma: float = 1.3
  categorical_dimensions: tuple[int, ...] | None = None
  categorical_indexes: tuple[int, ...] | None = None
  categorical_embedding_dimensions: tuple[int, ...] | None = None
  independent_glu_layers: int = 2
  shared_glu_layers: int = 2
  epsilon: float = 1e-15
  virtual_batch_size: int = 128
  momentum: float = 0.98
  mask_type: str = "sparsemax"
  model_data_type: jnp.dtype = jnp.float32

  def setup(self) -> None:
    """Sets up the module definition when it's called.

    Raises:
      ValueError: An error when successive_network_steps is less or equal to 0.
        Or when independent_glu_layers and shared_glu_layers are equal 0.
    """
    if self.successive_network_steps <= 0:
      raise ValueError("successive_network_steps should be a positive integer.")
    if self.independent_glu_layers == 0 and self.shared_glu_layers == 0:
      raise ValueError(
          "shared_glu_layers and independent_glu_layers can not be both zero."
      )

    categorical_dimensions = (
        [] if not self.categorical_dimensions else self.categorical_dimensions
    )
    categorical_indexes = (
        [] if not self.categorical_indexes else self.categorical_indexes
    )
    self.embedder = EmbeddingGenerator(
        self.input_dimensions,
        categorical_dimensions,
        categorical_indexes,
        self.categorical_embedding_dimensions,
        self.model_data_type,
    )
    self.post_embed_dim = self.embedder.post_embedding_dimension
    self.tabnet = TabNetCoreTraining(
        input_dimensions=self.post_embed_dim,
        output_dimensions=self.output_dimensions,
        prediction_layer_dimension=self.prediction_layer_dimension,
        attention_layer_dimension=self.attention_layer_dimension,
        successive_network_steps=self.successive_network_steps,
        attention_gamma=self.attention_gamma,
        independent_glu_layers=self.independent_glu_layers,
        shared_glu_layers=self.shared_glu_layers,
        epsilon=self.epsilon,
        virtual_batch_size=self.virtual_batch_size,
        momentum=self.momentum,
        mask_type=self.mask_type,
        model_data_type=self.model_data_type,
    )

  @nn.compact
  def __call__(
      self,
      *,
      input_x: jnp.ndarray,
      forward_masks: bool = False,
      train: bool = False,
  ) -> (
      tuple[list[jnp.ndarray], list[jnp.ndarray]]
      | tuple[jnp.ndarray, dict[int, jnp.ndarray]]
      | tuple[jnp.ndarray, jnp.ndarray, float]
  ):
    """Forward pass implementation for the layer.

    Args:
      input_x: An input array to be processed by the TabNet layer.
      forward_masks: An indicator if the forward_masks process should be called.
      train: An indicator whether the pass is during training or not.

    Returns:
      The output of the forward pass.
    """
    if not forward_masks:
      if input_x.dtype != self.model_data_type:
        input_x = jnp.array(input_x, dtype=self.model_data_type)
      embedder_output = self.embedder(input_x=input_x)
      return self.tabnet(input_x=embedder_output, train=train)
    embedder_output = self.embedder(input_x=input_x)
    return self.tabnet(input_x=embedder_output, train=train, forward_masks=True)


class TabNetAgent(base_agent.BaseAgent):
  """Reinforcement Learning agent based on TabNet architecture."""

  def build_flax_module(self):
    """Build the TabNet architecture."""
    return TabNet(
        input_dimensions=self.hyperparameters.input_dimensions,
        output_dimensions=self.hyperparameters.action_space,
        categorical_indexes=self.hyperparameters.categorical_indexes,
        categorical_dimensions=self.hyperparameters.categorical_dimensions,
        prediction_layer_dimension=self.hyperparameters.prediction_layer_dimension,
        attention_layer_dimension=self.hyperparameters.attention_layer_dimension,
        successive_network_steps=self.hyperparameters.successive_network_steps,
        attention_gamma=self.hyperparameters.attention_gamma,
        categorical_embedding_dimensions=self.hyperparameters.categorical_embedding_dimensions,
        independent_glu_layers=self.hyperparameters.independent_glu_layers,
        shared_glu_layers=self.hyperparameters.shared_glu_layers,
        epsilon=self.hyperparameters.epsilon,
        virtual_batch_size=self.hyperparameters.virtual_batch_size,
        momentum=self.hyperparameters.momentum,
        mask_type=self.hyperparameters.mask_type,
        model_data_type=self.hyperparameters.model_data_type,
    )



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of all the available optimizers."""
import dataclasses
from typing import Any, Callable

from ml_collections.config_dict import config_dict
import optax


@dataclasses.dataclass(frozen=True)
class _Optimizer:
  """Instance with an optimizer name and its instance.

  Attributes:
    name: An optimizer name.
    instance: An optimizer instance.
  """

  name: str
  instance: Callable[[Any], optax.GradientTransformation]


_ALL_OPTIMIZERS = config_dict.ConfigDict(
    dict(adam=_Optimizer("adam", optax.adam))
)


def get_optimizer(
    *, optimizer_name: str,
) -> Callable[[Any], optax.GradientTransformation]:
  """Maps the optimizer name with the corresponding optimizer function.

  Args:
    optimizer_name: Optimizer name.

  Returns:
    The requested optimizer function.

  Raises:
    LookupError: An error when trying to access an unavailable optimizer.
  """
  if optimizer_name not in _ALL_OPTIMIZERS:
    raise LookupError(f"Unrecognized optimizer name: {optimizer_name}")
  return _ALL_OPTIMIZERS[optimizer_name].instance



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all agents."""

import abc
import dataclasses
from typing import Any, Callable, Final, Mapping, Protocol
import flax
from flax.core import frozen_dict
from flax.training import train_state
import jax
import jax.numpy as jnp
from ml_collections.config_dict import config_dict

_EPSILON: Final[float] = 1e-8
FrozenDict = frozen_dict.FrozenDict


class CalculatePretrainReward(Protocol):

  def __call__(
      self, batch: jnp.ndarray, actions: jnp.ndarray, sign_rewards: bool
  ) -> jnp.ndarray:
    ...


class BaseAgentState(train_state.TrainState):
  """Adds batch_stats to the vanilla Flax TrainState class."""

  batch_stats: flax.core.FrozenDict | dict[str, Any]
  exploration_exploitation_epsilon: float


def apply_policy(
    *,
    parameters: FrozenDict | Mapping[str, Any],
    apply_function: Callable[..., Any],
    batch: jnp.ndarray,
    **kwargs,
) -> tuple[jnp.ndarray, ...]:
  """Applies policy on an input batch.

  Args:
    parameters: The current parameters of the trained agent (model).
    apply_function: An apply function of the trained agent (model).
      batch: An input data batch used to predict an action of shape (batch_size,
        input_dimensions).
    **kwargs: Other arguments, e.g. train and/or mutable.

  Returns:
    The output of an agent's apply_function.
  """
  if "batch_statistics" in kwargs:
    variables = {
        "params": parameters,
        "batch_stats": kwargs["batch_statistics"],
    }
    kwargs.pop("batch_statistics", None)
  else:
    variables = {"params": parameters}
  return apply_function(variables=variables, input_x=batch, **kwargs)


def predict_actions(
    *,
    exploration_exploitation_epsilon: float,
    prng_key: jax.Array,
    probabilities: jnp.ndarray,
) -> jnp.ndarray:
  """Predicts actions depending on the exploration/exploitation rate.

  Args:
    exploration_exploitation_epsilon: An exploration exploitation rate affecting
      if the most optimal actions is selected or not.
    prng_key: A PRNG key to ensure repeatability.
    probabilities: An array with each action probability of shape (,
      total_number_of_possible_actions, ).

  Returns:
    An array with predicted action of shape (1, 1).
  """
  normalized_probabilities = probabilities / jnp.linalg.norm(
      probabilities, 1, axis=0, keepdims=True
  )
  epsilon_adjusted_probabilities = (
      exploration_exploitation_epsilon / len(normalized_probabilities)
      + (1 - exploration_exploitation_epsilon) * normalized_probabilities
  )
  return jax.random.choice(
      key=prng_key, a=probabilities.shape[0], p=epsilon_adjusted_probabilities
  ).reshape(1, 1)


@dataclasses.dataclass(frozen=True)
class PredictionOutput:
  """A mapping of agent experience building blocks and respective arrays.

  Attributes:
    prediction_seed: A prediction seed used to select the action.
    state: An array with state (i.e. features) based on which an agent has to
      predict the most optimal decision. It's of shape (1, input_dimensions)
    action: An array with an action predicted by an agent based on the provided
      context of shape (1, 1).
    value: An array with a "critic" value of shape (1, 1).
    log_probability: An array of log probability of the predicted action of
      shape (1, 1).
    done: An array with booleans indicating if the end of a series of events (if
      applicable, e.g. for TabNet) of shape (1, 1).
    attentive_transformer_loss: An array with attentive transformer losses (if
      applicable, e.g. for TabNet) of shape (1, 1).
  """

  prediction_seed: int
  state: jnp.ndarray
  action: jnp.ndarray
  value: jnp.ndarray
  log_probability: jnp.ndarray
  done: jnp.ndarray
  attentive_transformer_loss: jnp.ndarray


class BaseAgent(metaclass=abc.ABCMeta):
  """Abstract parent class for all agent classes.

  Attributes:
    hyperparameters: The hyperparameteres for the agent class.
    flax_module: A Flax model used for training and making predictions.
  """

  def __init__(self, *, hyperparameters: config_dict.ConfigDict) -> None:
    """Initializes the BaseAgent class.

    Args:
      hyperparameters: Agent's hyperparameters.
    """
    self.hyperparameters = hyperparameters
    self.flax_module = self.build_flax_module()

  @abc.abstractmethod
  def build_flax_module(self) -> Any:
    """Builds a Flax module / model for training."""

  def get_dummy_inputs(self) -> jnp.ndarray:
    """Outputs dummy inputs to initialize a Flax module / model."""
    return jnp.zeros(
        (
            self.hyperparameters.batch_size,
            self.hyperparameters.input_dimensions,
        ),
        dtype=self.hyperparameters.model_data_type,
    )

  def predict(
      self,
      *,
      agent_state: BaseAgentState,
      batch: jnp.ndarray,
      prediction_seed: int,
  ) -> PredictionOutput:
    """Returns a prediction output (incl. actions) given an input batch.

    Args:
      agent_state: Current agent state.
      batch: An input data batch used to predict an action of shape (batch_size,
        input_dimensions).
      prediction_seed: A prediction seed used to select an action.
    """
    (log_probabilities, value, attentive_transformer_loss), _ = apply_policy(
        parameters=agent_state.params,
        batch_statistics=agent_state.batch_stats,
        apply_function=agent_state.apply_fn,
        batch=batch,
        train=True,
        mutable=["batch_stats"],
    )
    predicted_action = predict_actions(
        exploration_exploitation_epsilon=agent_state.exploration_exploitation_epsilon,
        prng_key=jax.random.PRNGKey(prediction_seed),
        probabilities=jnp.exp(log_probabilities).flatten(),
    )
    log_probabilities = jnp.take(log_probabilities, predicted_action)
    attentive_transformer_loss = jnp.asarray(
        attentive_transformer_loss
    ).reshape(1, 1)
    if self.hyperparameters.replace_nans_in_prediction:
      value = jnp.nan_to_num(value, nan=_EPSILON)
      log_probabilities = jnp.nan_to_num(log_probabilities, nan=_EPSILON)
      attentive_transformer_loss = jnp.nan_to_num(
          attentive_transformer_loss, nan=_EPSILON
      )
    return PredictionOutput(
        prediction_seed=prediction_seed,
        state=batch,
        action=predicted_action,
        value=value,
        log_probability=log_probabilities,
        done=jnp.repeat(True, 1).reshape(1, 1),
        attentive_transformer_loss=attentive_transformer_loss,
    )

  def evaluate_predict(
      self,
      *,
      agent_state: BaseAgentState,
      batch: jnp.ndarray,
  ) -> jnp.ndarray:
    """Returns predicted actions given an input batch.

    Args:
      agent_state: Current agent state.
      batch: An input data batch used to predict an action of shape (batch_size,
        input_dimensions).
    """
    log_probabilities, _, _ = apply_policy(
        parameters=agent_state.params,
        apply_function=agent_state.apply_fn,
        batch=batch,
        batch_statistics=agent_state.batch_stats,
        train=False,
        mutable=False,
    )
    return jnp.argmax(log_probabilities, axis=1).reshape(batch.shape[0], 1)

  def pretrain_predict(
      self,
      *,
      batch: jnp.ndarray,
      agent_state: BaseAgentState,
      calculate_pretrain_reward: CalculatePretrainReward,
      prediction_seed: int,
  ) -> jnp.ndarray:
    """Returns a prediction for a single observation during pre-training.

    Args:
      batch: An array with with a unique user id and user data.
      agent_state: The current agent state.
      calculate_pretrain_reward: A function that calculates (or predicts)
        rewards for actions taken during pre-training process.
      prediction_seed: A prediction seed used to select the action.

    Raises:
      ValueError: An error when the input batch isn't one-dimensional.
    """
    if batch.ndim != 1:
      raise ValueError(
          "The input batch must be one-dimensional. There are"
          f" {batch.ndim} dimensions now."
      )
    context = jnp.expand_dims(
        batch[: self.hyperparameters.input_dimensions], axis=0
    )
    if isinstance(prediction_seed, jnp.ndarray):
      prediction_seed = prediction_seed.reshape().astype(jnp.int32)
    pretraining_prediction_output = self.predict(
        agent_state=agent_state,
        batch=context,
        prediction_seed=prediction_seed,
    )
    reward = calculate_pretrain_reward(
        batch=batch,
        actions=pretraining_prediction_output.action,
        sign_rewards=self.hyperparameters.sign_rewards,
    )
    return jnp.concatenate(
        [
            pretraining_prediction_output.state,
            pretraining_prediction_output.action,
            reward,
            pretraining_prediction_output.value,
            pretraining_prediction_output.log_probability,
            pretraining_prediction_output.done,
            pretraining_prediction_output.attentive_transformer_loss,
        ],
        axis=1,
    ).flatten()



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all actions."""
import abc

from ml_collections.config_dict import config_dict

DEFAULT_HYPERPARAMETERS = config_dict.ConfigDict(
    dict(
        action_space=config_dict.placeholder(list),
        actions_name="base_actions",
    )
)


class BaseActions(metaclass=abc.ABCMeta):
  """Abstract parent class for all actions classes.

  Attributes:
    hyperparameters: The hyperparameteres for the action class.
  """

  def __init__(self, *, hyperparameters: config_dict.ConfigDict) -> None:
    """Initializes the BaseActions class.

    Args:
      hyperparameters: BaseActions'/ experiment hyperparameters.
    """
    self.hyperparameters = hyperparameters



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all rewards."""

import abc

import jax.numpy as jnp
from ml_collections.config_dict import config_dict
import tensorflow as tf

DEFAULT_HYPERPARAMETERS = config_dict.ConfigDict(
    dict(
        sign_rewards=True,
        reward_name="base_reward",
    )
)


class BaseReward(metaclass=abc.ABCMeta):
  """Abstract parent class for all rewards classes.

  Attributes:
    hyperparameters: A rewards class hyperparameters.
  """

  def __init__(self, *, hyperparameters: config_dict.ConfigDict) -> None:
    """Initalizes the BaseReward class.

    Args:
      hyperparameters: The hyperparameteres for the reward class.
    """
    self.hyperparameters = hyperparameters

  @abc.abstractmethod
  def calculate_reward(
      self, actions: tf.Tensor, reactions: tf.Tensor, sign_rewards: bool
  ) -> tf.Tensor:
    """Calculates rewards given the selected actions and end-user reactions."""

  @abc.abstractmethod
  def calculate_pretrain_reward(
      self, batch: jnp.ndarray, actions: jnp.ndarray, sign_rewards: bool
  ) -> jnp.ndarray:
    """Calculates pretrain rewards given the selected actions."""

  def calculate_evaluation_reward(
      self,
      *,
      predicted_actions: jnp.ndarray,
      target_actions: jnp.ndarray,
  ) -> jnp.ndarray:
    """Calculates evaluation rewards given the selected and target actions.

    Args:
      predicted_actions: A batch with selected actions to be evaluated of shape
        (batch_size, 1).
      target_actions: An array with target actions of shape (batch_size, 1).

    Returns:
      A batch with calculated evaluation rewards.
    """
    return jnp.where(predicted_actions == target_actions, 1.0, -1.0)



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all data pipelines."""
import abc
import functools
from typing import Callable, Final
from absl import logging
import jax
import jax.numpy as jnp
from ml_collections.config_dict import config_dict
import numpy as np
import tensorflow as tf

AUTOTUNE = tf.data.experimental.AUTOTUNE

DEFAULT_HYPERPARAMETERS = config_dict.ConfigDict(
    dict(
        input_dimensions=config_dict.placeholder(int),
        categorical_dimensions=config_dict.placeholder(list),
        categorical_indexes=config_dict.placeholder(list),
        shuffle_size=512,
        batch_size=1024,
        evaluation_batch_size=8,
        train_dataset_size=config_dict.placeholder(int),
        data_pipeline_name="base_data_pipeline",
        reactions_dimensions=1,
    )
)

_TRAIN: Final[str] = "train"
_PRETRAIN: Final[str] = "pretrain"
_EVALUATION: Final[str] = "evaluation"
_CalculateRewardFunction = Callable[[tf.Tensor, tf.Tensor, bool], tf.Tensor]


def process_data(
    *,
    tensor: tf.Tensor,
    split: str = "train",
    hyperparameters: config_dict.ConfigDict,
    reward_calculation_function: _CalculateRewardFunction | None = None,
) -> dict[str, tf.Tensor]:
  """Preprocesses tensors in either train or evaluation format.

  Args:
    tensor: A tensor from a TF Dataset.
    split: A split of data to preprocess.
    hyperparameters: A data pipeline class hyperparameters.
    reward_calculation_function: A function to calculate rewards based on taken
      action and recorded user reactions.

  Returns:
    A dictionary of tensors in either train or evaluation format.

  Raises:
    ValueError: An error when split not equal to "train" or "evaluation".
  """
  if split == _TRAIN:
    (
        states_tensor,
        actions_tensor,
        values_tensor,
        log_probabilities_tensor,
        dones_tensor,
        attentive_transformer_losses_tensor,
        reactions_tensor,
    ) = tf.split(
        tensor,
        [
            hyperparameters.input_dimensions,
            1,
            1,
            1,
            1,
            1,
            hyperparameters.reactions_dimensions,
        ],
    )
    rewards_tensor = reward_calculation_function(
        actions_tensor,
        reactions_tensor,
        hyperparameters.sign_rewards,
    )
    observation = dict(
        states=states_tensor,
        actions=actions_tensor,
        rewards=rewards_tensor,
        values=values_tensor,
        log_probabilities=log_probabilities_tensor,
        dones=dones_tensor,
        attentive_transformer_losses=attentive_transformer_losses_tensor,
    )
  elif split == _EVALUATION:
    states_tensor, target_actions_tensor = tf.split(
        tensor, [hyperparameters.input_dimensions, 1]
    )
    observation = dict(
        states=states_tensor,
        target_actions=target_actions_tensor,
    )
  else:
    raise ValueError(
        f"Split must be either 'train' or 'evaluation', you passed: {split!r}"
    )
  return observation


def calculate_train_and_evaluation_batch_size(
    *,
    process_count: int,
    number_of_devices: int,
    hyperparameters: config_dict.ConfigDict,
) -> tuple[int, int]:
  """Verifies and calculates train and evaluation batch sizes.

  Args:
    process_count: A number of processes detected by JAX.
    number_of_devices: A number of devices (e.g. TPUs) detected by JAX.
    hyperparameters: A data pipeline class hyperparameters.

  Returns:
    Train batch size per device and evaluation batch size per device.

  Raises:
    ValueError: An error when: 1) batch_size isn't divisible by the number of
    detected processes, 2) train_dataset_size / batch_size is less than
    train_steps, 3) evaluation_batch_size isn't divisible by the number of
    detected processes, 4) per_host_batch_size isn't divisible by the number of
    detected devices or 5) per_host_evaluation_batch_size isn't divisible by the
    number of detected devices.
  """
  if hyperparameters.batch_size % process_count:
    raise ValueError(
        f"process_count={process_count} must divide"
        f" batch_size={hyperparameters.batch_size}."
    )
  per_host_batch_size = hyperparameters.batch_size // process_count
  if per_host_batch_size % number_of_devices != 0:
    raise ValueError(
        f"number_of_devices={number_of_devices} must divide"
        f" per_host_batch_size={per_host_batch_size}."
    )
  if (
      hyperparameters.train_dataset_size / per_host_batch_size
      < hyperparameters.train_steps
  ):
    raise ValueError(
        f"train_dataset_size={hyperparameters.train_dataset_size} divided"
        f" by per_host_batch_size={per_host_batch_size} must be"
        f" more than train_steps={hyperparameters.train_steps}."
    )
  evaluation_batch_size = (
      hyperparameters.evaluation_batch_size or hyperparameters.batch_size
  )
  if evaluation_batch_size % process_count:
    raise ValueError(
        f"process_count={process_count} must divide"
        f" evaluation_batch_size={evaluation_batch_size}."
    )
  per_host_evaluation_batch_size = evaluation_batch_size // process_count
  if per_host_evaluation_batch_size % number_of_devices != 0:
    raise ValueError(
        f"number_of_devices={number_of_devices} must divide"
        f" per_host_evaluation_batch_size={per_host_evaluation_batch_size}."
    )
  return (
      per_host_batch_size,
      per_host_evaluation_batch_size,
  )


class BaseDataPipeline(metaclass=abc.ABCMeta):
  """Abstract parent class for all data pipeline classes.

  Attributes:
    hyperparameters: Training hyperparameters.
    per_host_batch_sizes: A sequence with per host train and evaluation batch
      sizes.
  """

  def __init__(
      self,
      *,
      hyperparameters: config_dict.ConfigDict,
  ) -> None:
    """Initializes the BaseDataPipeline class.

    Args:
      hyperparameters: The hyperparameteres for the data pipeline class.
    """
    self.hyperparameters = hyperparameters
    self._dataset_rng_key = jax.random.PRNGKey(
        hyperparameters.training_rng_seed
    )

  def build_tensorflow_pipeline(
      self,
      *,
      data: jnp.ndarray | np.ndarray,
      split: str = "train",
      batch_size: int = 1024,
      drop_remainder: bool = True,
      reward_calculation_function: _CalculateRewardFunction | None = None,
  ) -> tf.data.Dataset:
    """Returns a TensorFlow data pipeline.

    Args:
      data: Either train or evaluation data.
      split: A split of data to preprocess.
      batch_size: A batch size in the pipeline.
      drop_remainder: An indicator whether to drop remainder during batching.
      reward_calculation_function: A function to calculate rewards based on
        taken action and recorded user reactions.

    Raises:
      ValueError: An error when split not equal to "train" or "evaluation".
    """
    dataset = tf.data.Dataset.from_tensor_slices(data)
    if split == _TRAIN or split == _PRETRAIN:
      dataset = dataset.shuffle(
          self.hyperparameters.shuffle_size,
          seed=jax.random.bits(self._dataset_rng_key).item(),
      )
    elif split == _EVALUATION:
      pass
    else:
      raise ValueError(
          "Split must be either 'train', 'pretrain' or 'evaluation', you"
          f" passed: {split}"
      )
    if split != _PRETRAIN:
      partial_process_data = functools.partial(
          process_data,
          split=split,
          hyperparameters=self.hyperparameters,
          reward_calculation_function=reward_calculation_function,
      )
      dataset = dataset.map(lambda x: partial_process_data(tensor=x), AUTOTUNE)
    dataset = dataset.batch(
        batch_size=batch_size, drop_remainder=drop_remainder
    )
    return dataset.prefetch(AUTOTUNE)

  @functools.cached_property
  def per_host_batch_sizes(self) -> tuple[int, int]:
    """Returns train and evaluation per host batch sizes."""
    process_count = jax.process_count()
    logging.info("JAX counted %s process(es).", process_count)
    number_of_devices = jax.local_device_count()
    logging.info("JAX counted %s local device(s).", number_of_devices)
    per_host_batch_size, per_host_evaluation_batch_size = (
        calculate_train_and_evaluation_batch_size(
            process_count=process_count,
            number_of_devices=number_of_devices,
            hyperparameters=self.hyperparameters,
        )
    )
    logging.info(
        "The calculated per_host_batch_size is : %s", per_host_batch_size
    )
    logging.info(
        "The calculated per_host_evaluation_batch_size is : %s",
        per_host_evaluation_batch_size,
    )
    return per_host_batch_size, per_host_evaluation_batch_size

  def train_data_pipeline(
      self,
      *,
      train_data: jnp.ndarray | np.ndarray,
      reward_calculation_function: _CalculateRewardFunction,
  ) -> tf.data.Dataset:
    """Returns a train TensorFlow data pipeline.

    Args:
      train_data: Train data.
      reward_calculation_function: A TensorFlow function to calculate rewards
        from "actions" and "reactions".
    """
    return self.build_tensorflow_pipeline(
        data=train_data,
        split=_TRAIN,
        batch_size=self.per_host_batch_sizes[0],
        drop_remainder=True,
        reward_calculation_function=reward_calculation_function,
    )

  def pretrain_data_pipeline(
      self,
      *,
      pretrain_data: jnp.ndarray | np.ndarray,
  ) -> tf.data.Dataset:
    """Returns a pretrain TensorFlow data pipeline.

    Args:
      pretrain_data: Pretrain data.
    """
    return self.build_tensorflow_pipeline(
        data=pretrain_data,
        split=_PRETRAIN,
        batch_size=self.per_host_batch_sizes[0],
        drop_remainder=True,
    )

  def evaluation_data_pipeline(
      self, *, evaluation_data: jnp.ndarray | np.ndarray
  ) -> tf.data.Dataset:
    """Returns an evaluation TensorFlow data pipeline.

    Args:
      evaluation_data: Evaluation data.
    """
    return self.build_tensorflow_pipeline(
        data=evaluation_data,
        split=_EVALUATION,
        batch_size=self.per_host_batch_sizes[1],
        drop_remainder=False,
    )



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of all the available losses."""
import dataclasses
from typing import Any, Callable, Mapping

import jax
import jax.numpy as jnp
from ml_collections.config_dict import config_dict

from optimus.agent_lib import base_agent


@dataclasses.dataclass()
class Trajectories:
  """Trajectories from an input batch.

  Attributes:
    context_state: Context state array of shape (batch_size, input_dimensions).
    actions: Actions array of shape (batch_size, action_space_length).
    original_log_probabilities: Original log probabilities array of shape
      (batch_size, action_space_length).
    returns: Returns array of shape (batch_size, ).
    advantages: Advantages array of shape (batch_size, ).
    attentive_transformer_losses: Attentive transformer losses array of shape
      (batch_size, ).
  """

  context_state: jnp.ndarray
  actions: jnp.ndarray
  original_log_probabilities: jnp.ndarray
  returns: jnp.ndarray
  advantages: jnp.ndarray
  attentive_transformer_losses: jnp.ndarray


def process_batch(
    *,
    batch: jnp.ndarray,
    hyperparameters: config_dict.ConfigDict,
) -> Trajectories:
  """Process a batch with trajectories.

  Args:
    batch: An input batch.
    hyperparameters: Experiment hyperparameters.

  Returns:
    Trajectories dataclass with the processed input data array.
  """
  context_state = batch[:, 0 : hyperparameters.input_dimensions]
  if hyperparameters.action_space_length > 1:
    skip_indexes_length = (
        hyperparameters.input_dimensions + hyperparameters.action_space_length
    )
    actions = jnp.array(
        batch[:, hyperparameters.input_dimensions : skip_indexes_length],
        dtype=jnp.int32,
    )
    actions = jnp.array_split(
        actions, hyperparameters.action_space_length, axis=1
    )
  else:
    skip_indexes_length = hyperparameters.input_dimensions
    actions = jnp.array(
        batch[:, hyperparameters.input_dimensions], dtype=jnp.int32
    )
  original_log_probabilities = batch[:, skip_indexes_length + 1]
  returns = batch[:, skip_indexes_length + 2]
  advantages = batch[:, skip_indexes_length + 3]
  attentive_transformer_losses = batch[:, skip_indexes_length + 4]
  return Trajectories(
      context_state,
      actions,
      original_log_probabilities,
      returns,
      advantages,
      attentive_transformer_losses,
  )


def calculate_entropy_and_ratios(
    predicted_log_probabilities: jnp.ndarray,
    *,
    actions: jnp.ndarray,
    original_log_probabilities: jnp.ndarray,
    action_space_length: int,
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Calculate entropy and ratios.

  Computes a loss as a sum of 3 components: the negative of the PPO clipped
  surrogate objective, the value function loss and the negative of the entropy
  bonus. For more on PPO see:

  https://arxiv.org/abs/1707.06347

  The loss is then modified with lambda_sparse and m_losses, required for
  TabNet training.

  Args:
    predicted_log_probabilities: An array with log probabilities from the
      apply_policy method in the loss calculations process.
    actions: An array with selected actions.
    original_log_probabilities: An array with log probabilities of selected
      actions.
    action_space_length: The number of actions sets.

  Returns:
    A tuple where the first element is the entropy between action
    probabilities and log probabilities and the second element is
    the ratios between log probabilities of taken actions and original log
    probabilities array of shape (batch_size, ).
  """
  if action_space_length > 1:
    predicted_probabilities = jax.tree_util.tree_map(
        jnp.exp, predicted_log_probabilities
    )
    entropy = jax.tree_util.tree_map(
        lambda x, y: jnp.sum(-x * y, axis=1).mean(),
        predicted_probabilities,
        predicted_log_probabilities,
    )
    entropy = jnp.mean(jnp.array(entropy, dtype=jnp.float32), axis=0)
    log_probs_action_taken = [
        jax.vmap(lambda p, a: p[a])(sub_probs, sub_actions).flatten()
        for sub_probs, sub_actions in zip(predicted_log_probabilities, actions)
    ]
    ratios = jax.tree_util.tree_map(
        lambda x, y: jnp.exp(x - y),
        log_probs_action_taken,
        list(jnp.repeat(original_log_probabilities, action_space_length)),
    )
    ratios = jnp.mean(jnp.array(ratios, dtype=jnp.float32), axis=0)
  else:
    predicted_probabilities = jnp.exp(predicted_log_probabilities)
    entropy = jnp.array(
        jnp.sum(
            -predicted_probabilities * predicted_log_probabilities, axis=1
        ).mean(),
        dtype=jnp.float32,
    )
    log_probabilities_action_taken = jax.vmap(lambda lp, a: lp[a])(
        predicted_log_probabilities, actions
    )
    ratios = jnp.exp(
        log_probabilities_action_taken - original_log_probabilities
    )
  return (entropy, ratios)


def calculate_tabnet_proximal_policy_optimization_loss(
    parameters: base_agent.FrozenDict | Mapping[str, Any],
    batch_statistics: base_agent.FrozenDict | Mapping[str, Any],
    apply_function: Callable[..., Any],
    *,
    batch: jnp.ndarray,
    clip_parameters_coefficient: float,
    hyperparameters: config_dict.ConfigDict,
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Calculate loss and update gradients.

  Computes a loss as a sum of 3 components: the negative of the PPO clipped
  surrogate objective, the value function loss and the negative of the entropy
  bonus. For more on PPO see:

  https://arxiv.org/abs/1707.06347

  The loss is then modified with lambda sparse and attentive transformer loss,
  required for TabNet training.

  Args:
    parameters: The current parameters of the trained agent (model).
    batch_statistics: The current batch statistics of the trained agent (model).
    apply_function: An apply function of the trained agent (model).
    batch: An input batch.
    clip_parameters_coefficient: The clipping coefficient used to clamp ratios
      in loss function.
    hyperparameters: Experiment hyperparameters.

  Returns:
    A tuple where the first element is step loss and the second
    element is the updated batch stats.
  """
  processed_batch = process_batch(batch=batch, hyperparameters=hyperparameters)
  (
      predicted_log_probabilities,
      values,
      attentive_transformer_losses,
  ), updated_batch_stats = base_agent.apply_policy(
      parameters=parameters,
      apply_function=apply_function,
      batch=processed_batch.context_state,
      batch_statistics=batch_statistics,
      train=True,
      mutable=["batch_stats"],
  )
  values = values[:, 0]
  value_loss = jnp.mean(jnp.square(processed_batch.returns - values), axis=0)
  entropy, ratios = calculate_entropy_and_ratios(
      predicted_log_probabilities,
      actions=processed_batch.actions,
      original_log_probabilities=processed_batch.original_log_probabilities,
      action_space_length=hyperparameters.action_space_length,
  )
  advantages = (
      processed_batch.advantages - processed_batch.advantages.mean()
  ) / (processed_batch.advantages.std() + 1e-8)
  pg_loss = ratios * advantages
  clipped_loss = advantages * jax.lax.clamp(
      1.0 - clip_parameters_coefficient,
      ratios,
      1.0 + clip_parameters_coefficient,
  )
  proximal_policy_optimization_loss = -jnp.mean(
      jnp.minimum(pg_loss, clipped_loss), axis=0
  )
  final_loss = (
      proximal_policy_optimization_loss
      + hyperparameters.value_function_coefficient * value_loss
      - hyperparameters.entropy_coefficient * entropy
  )
  attentive_transformer_losses = jnp.mean(attentive_transformer_losses)
  final_loss = (
      final_loss - hyperparameters.lambda_sparse * attentive_transformer_losses
  )
  return final_loss, updated_batch_stats


@dataclasses.dataclass(frozen=True)
class _Loss:
  """Instance with a loss name and its instance.

  Attributes:
    name: A loss name.
    instance: A loss instance.
  """

  name: str
  instance: Any


_ALL_LOSSES = config_dict.FrozenConfigDict(
    dict(
        tabnet_proximal_policy_optimization_loss=_Loss(
            "tabnet_proximal_policy_optimization_loss",
            calculate_tabnet_proximal_policy_optimization_loss,
        )
    )
)


def get_loss(
    *,
    loss_name: str,
) -> Callable[[config_dict.ConfigDict], Callable[[float], float]]:
  """Maps the loss name with the corresponding loss function.

  Args:
    loss_name: A loss name.

  Returns:
    The requested loss function.

  Raises:
    LookupError: An error when trying to access an unavailable loss.
  """
  if loss_name not in _ALL_LOSSES:
    raise LookupError(f"Unrecognized loss name: {loss_name}")
  return _ALL_LOSSES[loss_name].instance



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of all the available schedules."""
import dataclasses
from typing import Any, Callable

from ml_collections.config_dict import config_dict
import optax


def check_schedule_hyperparameters(
    *,
    schedule_hyperparameters: config_dict.ConfigDict,
    expected_keys: tuple[str, ...],
    other_keys: tuple[str, ...] = ("schedule",),
) -> None:
  """Checks if all required schedule hyperparameters are present in the experiment dictionary.

  Args:
    schedule_hyperparameters: The schedule hyperparameters.
    expected_keys: A list with required schedule hyperparameters.
    other_keys: A tuple of keys that are part of schedule hyperparameters,
      but not required by the schedule function.

  Raises:
    ValueError: An error when schedule hyperparameters are invalid.
  """
  expected_keys = expected_keys + other_keys
  if set(schedule_hyperparameters.keys()) != set(expected_keys):
    raise ValueError(
        "Provided schedule_hyperparameters keys are invalid. Received:"
        f" {sorted(schedule_hyperparameters)}, Expected:"
        f" {sorted(expected_keys)}"
    )


def linear_schedule(
    schedule_hyperparameters: config_dict.ConfigDict,
) -> Callable[[Any], Any]:
  """Returns a linear schedule function.

  Args:
    schedule_hyperparameters: The schedule hyperparameters.
  """
  expected_keys = tuple(["initial_value", "end_value", "transition_steps"])
  check_schedule_hyperparameters(
      schedule_hyperparameters=schedule_hyperparameters,
      expected_keys=expected_keys,
  )
  linear_schedule_hyperparameters = {
      "init_value": schedule_hyperparameters.initial_value,
      "end_value": schedule_hyperparameters.end_value,
      "transition_steps": schedule_hyperparameters.transition_steps,
  }
  return optax.linear_schedule(**linear_schedule_hyperparameters)


def constant_schedule(
    schedule_hyperparameters: config_dict.ConfigDict,
) -> Callable[[Any], Any]:
  """Returns a constant schedule function.

  Args:
    schedule_hyperparameters: The schedule hyperparameters.
  """
  expected_keys = tuple(["initial_value"])
  check_schedule_hyperparameters(
      schedule_hyperparameters=schedule_hyperparameters,
      expected_keys=expected_keys,
  )
  return optax.constant_schedule(
      value=schedule_hyperparameters.initial_value
  )


def warmup_exponential_decay_schedule(
    schedule_hyperparameters: config_dict.ConfigDict,
) -> Callable[[Any], Any]:
  """Returns a warmup exponential decay schedule schedule function.

  Args:
    schedule_hyperparameters: The schedule hyperparameters.
  """
  expected_keys = tuple([
      "initial_value",
      "peak_value",
      "warmup_steps",
      "transition_steps",
      "decay_rate",
      "transition_begin",
      "staircase",
      "end_value",
  ])
  check_schedule_hyperparameters(
      schedule_hyperparameters=schedule_hyperparameters,
      expected_keys=expected_keys,
  )
  warmup_exponential_decay_schedule_hyperparameters = {
      "init_value": schedule_hyperparameters.initial_value,
      "peak_value": schedule_hyperparameters.peak_value,
      "warmup_steps": schedule_hyperparameters.warmup_steps,
      "transition_steps": schedule_hyperparameters.transition_steps,
      "decay_rate": schedule_hyperparameters.decay_rate,
      "transition_begin": schedule_hyperparameters.transition_begin,
      "staircase": schedule_hyperparameters.staircase,
      "end_value": schedule_hyperparameters.end_value,
  }
  return optax.warmup_exponential_decay_schedule(
      **warmup_exponential_decay_schedule_hyperparameters
  )


@dataclasses.dataclass(frozen=True)
class _Schedule:
  """Instance with a schedule name and its instance.

  Attributes:
    name: A schedule name.
    instance: A schedule instance.
  """

  name: str
  instance: Any


_ALL_SCHEDULES = config_dict.FrozenConfigDict(
    dict(
        linear_schedule=_Schedule("linear_schedule", linear_schedule),
        constant_schedule=_Schedule("constant_schedule", constant_schedule),
        warmup_exponential_decay_schedule=_Schedule(
            "warmup_exponential_decay_schedule",
            warmup_exponential_decay_schedule,
        ),
    ),
)


def get_schedule(
    *,
    schedule_name: str,
) -> Callable[[config_dict.ConfigDict], Callable[[float], float]]:
  """Maps the schedule name with the corresponding schedule function.

  Args:
    schedule_name: Schedule name.

  Returns:
    The requested schedule function.

  Raises:
    LookupError: An error when trying to access an unavailable schedule.
  """
  if schedule_name not in _ALL_SCHEDULES:
    raise LookupError(f"Unrecognized schedule name: {schedule_name}")
  return _ALL_SCHEDULES[schedule_name].instance



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of all the available data pipelines."""
import dataclasses

from ml_collections.config_dict import config_dict

from optimus.data_pipeline_lib import base_data_pipeline


@dataclasses.dataclass(frozen=True)
class _DataPipeline:
  """Instance with and a data pipeline class name, its instance and hyperparameters.

  Attributes:
    name: A name of a data pipeline class.
    instance: An instance of a data pipeline class.
    hyperparameters: Hyperparameters of a data pipeline class.
  """

  name: str
  instance: type[base_data_pipeline.BaseDataPipeline]
  hyperparameters: config_dict.ConfigDict


_ALL_DATA_PIPELINES = config_dict.ConfigDict(
    dict(
        base_data_pipeline=_DataPipeline(
            "base_data_pipeline",
            base_data_pipeline.BaseDataPipeline,
            base_data_pipeline.DEFAULT_HYPERPARAMETERS,
        )
    )
)


def get_data_pipeline(
    *, data_pipeline_name: str = "base_data_pipeline"
) -> type[base_data_pipeline.BaseDataPipeline]:
  """Maps the data pipeline name with the corresponding data pipeline class.

  Args:
    data_pipeline_name: Data pipeline name.

  Returns:
    The requested data pipeline class.

  Raises:
    LookupError: An error when trying to access an unavailable data pipeline
    class.
  """
  if data_pipeline_name not in _ALL_DATA_PIPELINES:
    raise LookupError(
        f"Unrecognized data pipeline class name: {data_pipeline_name}"
    )
  return _ALL_DATA_PIPELINES[data_pipeline_name].instance


def get_data_pipeline_hyperparameters(
    *, data_pipeline_name: str = "base_data_pipeline"
) -> config_dict.ConfigDict:
  """Maps the data pipeline name with the corresponding data pipeline hyperparameters.

  Args:
    data_pipeline_name: Data pipeline name.

  Returns:
    The requested data pipeline class hyperparameters.

  Raises:
    LookupError: An error when trying to access an unavailable data pipeline
    class.
  """
  if data_pipeline_name not in _ALL_DATA_PIPELINES:
    raise LookupError(
        f"Unrecognized data pipeline class name: {data_pipeline_name}"
    )
  return _ALL_DATA_PIPELINES[data_pipeline_name].hyperparameters



# Copyright 2023 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of all the available actions."""
import dataclasses

from ml_collections.config_dict import config_dict

from optimus.actions_lib import base_actions


@dataclasses.dataclass(frozen=True)
class _Actions:
  """Instance with and an actions class name, its instance and hyperparameters.

  Attributes:
    name: A name of an actions class.
    instance: An instance of an actions class.
    hyperparameters: Hyperparameters of an actions class.
  """

  name: str
  instance: type[base_actions.BaseActions]
  hyperparameters: config_dict.ConfigDict


_ALL_ACTIONS = config_dict.ConfigDict(
    dict(
        base_actions=_Actions(
            "base_actions",
            base_actions.BaseActions,
            base_actions.DEFAULT_HYPERPARAMETERS,
        )
    )
)


def get_actions(
    *, actions_name: str = "base_actions"
) -> type[base_actions.BaseActions]:
  """Maps the actions name with the corresponding actions class.

  Args:
    actions_name: Actions name.

  Returns:
    The requested actions class.

  Raises:
    LookupError: An error when trying to access an unavailable actions class.
  """
  if actions_name not in _ALL_ACTIONS:
    raise LookupError(f"Unrecognized actions class name: {actions_name}")
  return _ALL_ACTIONS[actions_name].instance


def get_actions_hyperparameters(
    *, actions_name: str = "base_actions"
) -> config_dict.ConfigDict:
  """Maps the actions name with the corresponding actions hyperparameters.

  Args:
    actions_name: Actions name.

  Returns:
    The requested actions class hyperparameters.

  Raises:
    LookupError: An error when trying to access an unavailable actions class.
  """
  if actions_name not in _ALL_ACTIONS:
    raise LookupError(f"Unrecognized actions class name: {actions_name}")
  return _ALL_ACTIONS[actions_name].hyperparameters



# Copyright 2024 Google LLC.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.mit.edu/~amini/LICENSE.md
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for the train operation."""
import itertools
import json
import os
from typing import Any, Final, Mapping, Sequence

from absl import logging
import jax
import jax.numpy as jnp
from ml_collections.config_dict import config_dict
import portpicker
import tensorflow as tf

_CPU: Final[str] = "cpu"
_GPU: Final[str] = "gpu"
_TPU: Final[str] = "tpu"
_VALID_ACCELERATORS = (_CPU, _GPU, _TPU)
_NESTED_HYPERPARAMETERS = (
    "optimizer_hyperparameters",
    "learning_rate_hyperparameters",
    "exploration_exploitation_hyperparameters",
)


def _verify_action_hyperparameters(
    *,
    hyperparameters: config_dict.ConfigDict,
) -> config_dict.ConfigDict:
  """Verifies if all the required action hyperparameters are present.

  Args:
    hyperparameters: Experiment hyperparameters.

  Returns:
    Verified experiment hyperparameters.
  """
  action_space = hyperparameters.action_space
  single_task_training = len(action_space) == 1
  with hyperparameters.unlocked():
    del hyperparameters["action_space"]
    hyperparameters["action_space"] = (
        action_space[0] if single_task_training else tuple(action_space)
    )
    hyperparameters["action_space_length"] = (
        1 if single_task_training else len(action_space)
    )
  return hyperparameters


def _verify_trainer_hyperparameters(
    *,
    hyperparameters: config_dict.ConfigDict,
) -> config_dict.ConfigDict:
  """Verifies if all the required trainer hyperparameters are present."""
  if "train_steps" not in hyperparameters:
    raise ValueError("train_steps hyperparameter missing.")
  return hyperparameters


def _aggregate_core_hyperparameters(
    *,
    actions_hyperparameters: config_dict.ConfigDict,
    agent_hyperparameters: config_dict.ConfigDict,
    reward_hyperparameters: config_dict.ConfigDict,
    data_pipeline_hyperparameters: config_dict.ConfigDict,
    trainer_hyperparameters: config_dict.ConfigDict,
) -> tuple[dict[str, Any], ...]:
  """Aggregates core hyperparameters.

  Args:
    actions_hyperparameters: A mapping between action class hyperparameters and
      values.
    agent_hyperparameters: A mapping between agent class hyperparameters and
      values.
    reward_hyperparameters: A mapping between reward class hyperparameters and
      values.
    data_pipeline_hyperparameters: A mapping between data pipeline class
      hyperparameters and values.
    trainer_hyperparameters: A mapping between trainer class hyperparameters and
      values.

  Returns:
    Aggregated core hyperparameters.
  """
  hyperparameters_sets = (
      actions_hyperparameters,
      agent_hyperparameters,
      reward_hyperparameters,
      data_pipeline_hyperparameters,
      trainer_hyperparameters,
  )
  return tuple(
      hyperparameter_set.to_dict()
      for hyperparameter_set in hyperparameters_sets
  )


def _merge_hyperparameters(
    *,
    hyperparameters: Sequence[Mapping[str, Any]],
) -> config_dict.ConfigDict:
  """Returns a merged version of hyperparameter mappings.

  Args:
    hyperparameters: A sequence of hyperparameter mappings to be merged.

  Raises:
    ValueError: An error when there's an overlap between merged hyperparameter
      dictionaries.
  """
  merged_hyperparameters = {}
  merged_hyperparameters_keys = []
  for hyperparameter_subset in hyperparameters:
    hyperparameter_subset_keys = list(hyperparameter_subset.keys())
    overlapping_hyperparameters = tuple([
        i
        for i in hyperparameter_subset_keys
        if i in merged_hyperparameters_keys
    ])
    if overlapping_hyperparameters:
      raise ValueError(
          "There is an overlap in the provided hyperparameters. The repeated"
          f" hyperparameters are: {overlapping_hyperparameters}"
      )
    merged_hyperparameters.update(hyperparameter_subset)
    merged_hyperparameters_keys += hyperparameter_subset_keys
  return config_dict.ConfigDict(merged_hyperparameters).lock()


def _convert_nested_hyperparameters(
    *,
    hyperparameters: config_dict.ConfigDict,
) -> config_dict.ConfigDict:
  """Converts all nested hyperparameters to the desired format.

  Args:
    hyperparameters: All the experiment hyperparameters.

  Returns:
    Experiment hyperparameters with converted the optimizer and learning rate
    hyperparameters values.
  """
  for key in _NESTED_HYPERPARAMETERS:
    if key not in hyperparameters:
      with hyperparameters.unlocked():
        hyperparameters[key] = config_dict.ConfigDict()
  for key in _NESTED_HYPERPARAMETERS:
    hyperparameters[key].unlock()
  return hyperparameters


def _update_hyperparameters_from_overrides(
    *,
    hyperparameters: config_dict.ConfigDict,
    manual_overrides: str | None = None,
    file_overrides: str | None = None,
) -> config_dict.ConfigDict:
  """Updates the experiment hyperparameters with hyperparameter overrides.

  Args:
    hyperparameters: The unmodified experiment hyperparameters.
    manual_overrides: A JSON string with hyperparameters overrides.
    file_overrides: A path to a JSON file with hyperparameter overrides.

  Returns:
    Experiment hyperparmeters with updated values.
  """
  hyperparameters_overrides_manual = (
      json.loads(manual_overrides) if manual_overrides else None
  )
  if file_overrides:
    with tf.io.gfile.GFile(file_overrides, "r") as f:
      hyperparameters_overrides_from_file = json.load(f)
  else:
    hyperparameters_overrides_from_file = None
  aggregated_hyperparameters_overrides = tuple(
      itertools.filterfalse(
          lambda overrides: not overrides,
          [
              hyperparameters_overrides_manual,
              hyperparameters_overrides_from_file,
          ],
      )
  )
  hyperparameters_overrides = _merge_hyperparameters(
      hyperparameters=aggregated_hyperparameters_overrides
  )
  for key in _NESTED_HYPERPARAMETERS:
    if key in hyperparameters_overrides:
      hyperparameters[key] = {}
  hyperparameters.update(hyperparameters_overrides)
  return hyperparameters


def build_hyperparameters(
    *,
    actions_hyperparameters: config_dict.ConfigDict,
    agent_hyperparameters: config_dict.ConfigDict,
    reward_hyperparameters: config_dict.ConfigDict,
    data_pipeline_hyperparameters: config_dict.ConfigDict,
    trainer_hyperparameters: config_dict.ConfigDict,
    checkpoint_directory: str,
    artifact_directory: str,
    hyperparameters_file: str | None = None,
    hyperparameters_overrides: str | None = None,
) -> config_dict.ConfigDict:
  """Assembles all the experiment hyperparameters.

  Args:
    actions_hyperparameters: A mapping between action class hyperparameters and
      values.
    agent_hyperparameters: A mapping between agent class hyperparameters and
      values.
    reward_hyperparameters: A mapping between reward class hyperparameters and
      values.
    data_pipeline_hyperparameters: A mapping between data_pipeline class
      hyperparameters and values.
    trainer_hyperparameters: A mapping between trainer class hyperparameters and
      values.
    checkpoint_directory: A path to the directory where model checkpoints
      will be saved.
    artifact_directory: A path to the directory where all the training artifacts
      (e.g. TensorBoard logs) will be saved.
    hyperparameters_file: A path to the file with hyperparameters overrides.
    hyperparameters_overrides: A JSON string with hyperparameters overrides.

  Returns:
    Assembled experiment hyperparameters.
  """
  core_hyperparameters = _aggregate_core_hyperparameters(
      actions_hyperparameters=actions_hyperparameters,
      agent_hyperparameters=agent_hyperparameters,
      reward_hyperparameters=reward_hyperparameters,
      data_pipeline_hyperparameters=data_pipeline_hyperparameters,
      trainer_hyperparameters=trainer_hyperparameters,
  )
  experiment_hyperparameters = core_hyperparameters + (
      {
          "checkpoint_directory": checkpoint_directory,
          "artifact_directory": artifact_directory,
      },
  )
  experiment_hyperparameters = _merge_hyperparameters(
      hyperparameters=experiment_hyperparameters
  )
  experiment_hyperparameters = _convert_nested_hyperparameters(
      hyperparameters=experiment_hyperparameters
  )
  experiment_hyperparameters = _update_hyperparameters_from_overrides(
      hyperparameters=experiment_hyperparameters,
      manual_overrides=hyperparameters_overrides,
      file_overrides=hyperparameters_file,
  )
  experiment_hyperparameters = _verify_action_hyperparameters(
      hyperparameters=experiment_hyperparameters
  )
  return _verify_trainer_hyperparameters(
      hyperparameters=experiment_hyperparameters
  )


def _get_model_data_type_from_string(*, model_data_type: str) -> jnp.dtype:
  """Maps a string with a jax.numpy dtype."""
  if model_data_type not in ("float32", "float64"):
    raise ValueError(f"Invalid dtype: {model_data_type}")
  return jnp.dtype(model_data_type)


def set_hyperparameters(
    *,
    actions_hyperparameters: config_dict.ConfigDict,
    agent_hyperparameters: config_dict.ConfigDict,
    reward_hyperparameters: config_dict.ConfigDict,
    data_pipeline_hyperparameters: config_dict.ConfigDict,
    trainer_hyperparameters: config_dict.ConfigDict,
    checkpoint_directory: str,
    artifact_directory: str,
    hyperparameters_file: str | None = None,
    hyperparameters_overrides: str | None = None,
) -> config_dict.ConfigDict:
  """Assembles all the experiment hyperparameters and saves them in a JSON file.

  Args:
    actions_hyperparameters: A mapping between action class hyperparameters and
      values.
    agent_hyperparameters: A mapping between agent class hyperparameters and
      values.
    reward_hyperparameters: A mapping between reward class hyperparameters and
      values.
    data_pipeline_hyperparameters: A mapping between data pipeline class
      hyperparameters and values.
    trainer_hyperparameters: A mapping between trainer class hyperparameters and
      values.
    checkpoint_directory: A path to the directory where model checkpoints
      will be saved.
    artifact_directory: A path to the directory where all the training artifacts
      (e.g. TensorBoard logs) will be saved.
    hyperparameters_file: A path to the file with hyperparameters overrides.
    hyperparameters_overrides: A JSON string with hyperparameters overrides.

  Returns:
    Assembled experiment hyperparameters.
  """
  hyperparameters = build_hyperparameters(
      actions_hyperparameters=actions_hyperparameters,
      agent_hyperparameters=agent_hyperparameters,
      reward_hyperparameters=reward_hyperparameters,
      data_pipeline_hyperparameters=data_pipeline_hyperparameters,
      trainer_hyperparameters=trainer_hyperparameters,
      checkpoint_directory=checkpoint_directory,
      artifact_directory=artifact_directory,
      hyperparameters_file=hyperparameters_file,
      hyperparameters_overrides=hyperparameters_overrides,
  )
  logging.info("Experiment hyperparameters are: %s", hyperparameters)
  hyperparameters_path = os.path.join(
      hyperparameters.artifact_directory, "hyperparameters.json"
  )
  with tf.io.gfile.GFile(hyperparameters_path, "w") as f:
    json.dump(hyperparameters.to_dict(), f)
    logging.info(
        "A JSON file with experiment hyperparameters saved as %s",
        hyperparameters_path,
    )
  if hyperparameters.model_data_type:
    converted_model_dtype = _get_model_data_type_from_string(
        model_data_type=hyperparameters.model_data_type
    )
    with hyperparameters.unlocked():
      del hyperparameters["model_data_type"]
      hyperparameters["model_data_type"] = converted_model_dtype
  return hyperparameters


def set_hardware(
    *,
    accelerator: str = "cpu",
    coordinator_address: str | None = None,
    process_number: int = 1,
    process_id: int = 0,
):
  """Sets up hardaware for the training loop.

  Args:
    accelerator: Machine accelerator type.
    coordinator_address: The IP address of process `0` and a port on which that
      process should launch a coordinator service.
    process_number: Number of processes.
    process_id: The ID number of the current process.

  Raises:
    ValueError: An error when using a GPU but not providing any coordinator
      address. Or when providing an incompatible accelerator name.
  """
  tf.config.experimental.set_visible_devices([], "GPU")
  logging.info(
      "Hid any GPU devices from Tensorflow since we only use it for data"
      " preprocessing."
  )
  if accelerator not in _VALID_ACCELERATORS:
    raise ValueError(
        f"Accelerator must be one of: {_VALID_ACCELERATORS}. "
        f"Received: {accelerator!r}"
    )
  if accelerator == "tpu":
    jax.distributed.initialize()
  elif accelerator == "gpu":
    if not coordinator_address:
      raise ValueError("coordinator_address is required when you use a GPU.")
    jax.distributed.initialize(
        coordinator_address, num_processes=process_number, process_id=process_id
    )
  elif accelerator == "cpu":
    port = portpicker.pick_unused_port()
    jax.distributed.initialize(
        f"localhost:{port}", num_processes=process_number, process_id=process_id
    )
  logging.info(
      "Initialized JAX distributed for a single-process setting.",
  )
  logging.info(
      "There are %s accelerator devices in the cluster visible to JAX.",
      jax.device_count(),
  )
  logging.info(
      (
          "There are %s accelerator devices attached to this host and visible"
          " to JAX."
      ),
      jax.local_device_count(),
  )
